using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;
using RegexManipulator;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Runtime.Serialization.Formatters.Binary;
using System.Runtime.Serialization;

namespace GTG
{
	public class Drawing
	{
		#region Constants

		public const float OuterRadius = 30; //R
		public const float OuterRadiusSquared = OuterRadius * OuterRadius; //R
		public const float InnerRadius = 26; //r
		public const float BezierControlPointDistance = OuterRadius + 60;
		public const float EdgeVerticalComponent = 8; //h
		public static readonly float EdgeHorizontalComponent; //e
		public const float EdgeVerticalField = 9;
		public const float LabelVerticalSpacing = 5; //lh
		public const float ArrowHeadHorizontalComponent = 7;
		public const float ArrowHeadVerticalComponent = 7;
		public const float StartStateIncomingEdgeLength = 40;
		const float PlusSignPolarAngle = (float)(-3.0 * Math.PI / 4.0);//3pi/4 Note: the angle
		//may sined and cosined at compile time.
		//Hence, only arithmetic and sqrt ops
		//are required.
		const float PlusSignEpsilon = 0;
		public static readonly Vector2F PlusSignVector;
		public const float MinusSignVerticalOffset = 0;
		public static readonly float StateLabelWidth = 2 * (InnerRadius - 1);

		public const string EpsilonLabel = "\x03B5";
		public const string OtherLabel = "other";
		public const string PlusString = "+";
		public const string MinusString = "-";

		public static readonly Vector2F BezierDirection1 = new Vector2F(1, -2).NormalizedVetor;
		public static readonly Vector2F BezierDirection2 = new Vector2F(-1, -2).NormalizedVetor;

		public static readonly Vector2F BV1 = OuterRadius * BezierDirection1;
		public static readonly Vector2F BV2 = BezierControlPointDistance * BezierDirection1;
		public static readonly Vector2F BV3 = BezierControlPointDistance * BezierDirection2;
		public static readonly Vector2F BV4 = OuterRadius * BezierDirection2;

		public static readonly float SelfEdgeLabelMaxWidth = BezierControlPointDistance
									* (BezierDirection1 - BezierDirection2).Magnitude + 50;
		#endregion

		#region Drawing Tools

		public readonly static Pen OuterCirclePen = new Pen(Color.Black, 2);
		public readonly static Pen InnerCirclePen = new Pen(Color.Blue, 1);
		public readonly static Brush PlusBrush = InnerCirclePen.Brush;
		public readonly static Brush MinusBrush = new SolidBrush(Color.FromArgb(128, 0, 128));
		public readonly static Brush InvalidRegexBrush = Brushes.Red;
		public readonly static Brush ValidRegexBrush = Brushes.Black;
		public readonly static Brush EpsilonRegexBrush = Brushes.Blue;
		public readonly static Brush OtherRegexBrush = new SolidBrush(Color.FromArgb(105, 139, 71));
		public readonly static Pen StartStateIncomingEdgePen = new Pen(MinusBrush);
		public readonly static Pen EdgePen = StartStateIncomingEdgePen;//new Pen(Color.Black, 2);
		public readonly static Pen SelectedEdgePen = new Pen(Color.DarkKhaki, 3);
		public readonly static Pen SelectedStateOuterCirclePen = new Pen(Color.DarkKhaki, 3);
		public readonly static Brush SelectedStateLabelBrush = Brushes.Brown;
		public readonly static Brush StateLabelBrush = Brushes.Gray;
		public readonly static Font EdgeLabelFont = new Font(FontFamily.GenericSansSerif, 14);//new Font("Century Gothic", 13, FontStyle.Regular);
		public readonly static Font StateLabelFont = new Font("Arial", 10, FontStyle.Bold);
		public readonly static Font PlusMinusFont = new Font("New Courier", 9, FontStyle.Regular);
		public readonly static Pen SelfEdgeSelectionPen = new Pen(Color.White, 2 * EdgeVerticalField);

		#endregion

		static Drawing()
		{

			//EdgePen.EndCap = LineCap.ArrowAnchor;

			EdgeHorizontalComponent = (float)Math.Sqrt(
				OuterRadius * OuterRadius - EdgeVerticalComponent * EdgeVerticalComponent);

			PlusSignVector = Vector2F.FromRadian(InnerRadius - PlusSignEpsilon,
											PlusSignPolarAngle);
		}
	}
	[Serializable]
	public enum EdgeType
	{
		/// <summary>
		/// The edge is labelled by an ordinary regular expression (may be epsilon)
		/// </summary>
		Labelled,
		/// <summary>
		/// The edge is an espilon transition
		/// </summary>
		Epsilon,
		/// <summary>
		/// The edge is labelled by other. This is equivalent to [~c1....cN] where ci's belong
		/// to the first sets of the other edges going out of that state. Many edges out of the
		/// same state may be labelled by other in a nondeterministic manner.
		/// </summary>
		Other
	}
	[Serializable]
	public enum LabelStatus
	{
		Epsilon,
		ValidRegex,
		InvalidRegex
	}
	[Serializable]
	public class EdgeGeometry
	{
		/// <summary>
		/// 
		/// </summary>
		public Point2F a;
		/// <summary>
		/// 
		/// </summary>
		public Point2F b;
		public PointF ArrowHeadTopPoint;
		public PointF ArrowHeadBottomPoint;
		/// <summary>
		/// The unit vector from p1 to p2 (the orientation of the label).
		/// If p1 and p2 collide, v is the zero vector
		/// </summary>
		public Vector2F v;
		/// <summary>
		/// The (-ve) unit vector perpendicular to p1-->p2 (the orientation of the label).
		/// If p1 and p2 collide, vperp is the zero vector
		/// </summary>
		public Vector2F vperp;

		/// <summary>
		/// The point that lies at the center of the edge label
		/// </summary>
		//			public Point2F raisedMidPoint;

		public float magnitude;
		/// <summary>
		/// After we take a number of units equal twice EdgeHorizontalComponent from
		/// the distance between the two states, signedRemainingDistance indicates the amount
		/// of distance remaining or needed.
		/// </summary>
		public float SignedRemainingDistance
		{
			get
			{
				return magnitude - 2 * Drawing.EdgeHorizontalComponent;
			}
		}

		public bool IsLeftToRight
		{
			get
			{
				return v.x > 0 || v.x == 0 && v.y >= 0;
			}
		}
		public EdgeGeometry()
		{
		}
		public EdgeGeometry(State src, State dst)
		{
			Reset(src.Location, dst.Location);
		}
		public void Reset(State src, State dst)
		{
			Reset(src.Location, dst.Location);
		}
		public void Reset(Point2F src, Point2F dst)
		{
			Vector2F V = dst - src;
			magnitude = V.Magnitude;
			if (magnitude == 0)
			{
				a = b = src.Clone();
				a.Offset(Drawing.EdgeHorizontalComponent, Drawing.EdgeVerticalComponent);
				//Note p2 has also been affected by the offset
				b = a.Clone(); //p1 and p2 should reference distinct objects for better use
				v = new Vector2F(0, 0);
				vperp = v.Clone();
				ArrowHeadTopPoint = ArrowHeadBottomPoint = new PointF();
				return;
			}
			v = 1 / magnitude * V; //normalize
			vperp = -1 * v.PerpendicularVector; //perp
			Vector2F usrc = Drawing.EdgeHorizontalComponent * v
							+ Drawing.EdgeVerticalComponent * vperp;
			Vector2F udst = -Drawing.EdgeHorizontalComponent * v
							+ Drawing.EdgeVerticalComponent * vperp;
			a = src + usrc;
			b = dst + udst;
			ArrowHeadTopPoint = b + Drawing.ArrowHeadVerticalComponent * vperp
								- Drawing.ArrowHeadHorizontalComponent * v;
			ArrowHeadBottomPoint = b - Drawing.ArrowHeadVerticalComponent * vperp
								- Drawing.ArrowHeadHorizontalComponent * v;
		}
		public Matrix EdgeLabelRotationMatrix
		{
			get
			{
				Matrix m;
				if (IsLeftToRight)
				{
					m = new Matrix(
						v.x, v.y,	//c, s
						-v.y, v.x,	//-s, c
						0, 0);
				}
				else
				{
					//angle = theta + PI
					m = new Matrix(
						-v.x, -v.y,
						v.y, -v.x,
						0, 0);
				}
				return m;
			}
		}
		public Point2F MidPoint
		{
			get
			{
				return .5f * (b - a) + a;
			}
		}
		public void Draw(Graphics g, string label, Font font, Brush brush, bool selected)
		{
			Matrix rotationMatrix;
			Matrix transformBackup;

			int labelAllowedWidth;
			SizeF labelSize;
			Point2F origin;
			StringFormat nowrap = new StringFormat(StringFormatFlags.NoWrap);


			labelAllowedWidth = (int)Math.Round(SignedRemainingDistance);
			if (labelAllowedWidth <= 0) //States are intersecting
				return;

			//Draw the edge
			if (selected)
			{
				g.DrawLine(Drawing.SelectedEdgePen, a, b);
				//Draw the arrow head
				g.DrawLine(Drawing.SelectedEdgePen, b, ArrowHeadTopPoint);
				g.DrawLine(Drawing.SelectedEdgePen, b, ArrowHeadBottomPoint);
			}
			else
			{
				g.DrawLine(Drawing.EdgePen, a, b);
				//Draw the arrow head
				g.DrawLine(Drawing.EdgePen, b, ArrowHeadTopPoint);
				g.DrawLine(Drawing.EdgePen, b, ArrowHeadBottomPoint);
			}

			//Prepare for drawing the label
			transformBackup = g.Transform.Clone(); //back the transformation up

			//Find the origin of rotation
			origin = MidPoint + Drawing.LabelVerticalSpacing * vperp; //not yet the origin

			//Measure label size
			labelSize = g.MeasureString(
				label,
				font,
				labelAllowedWidth,
				nowrap);

			//Now translate the origin based on the edge's direction
			if (IsLeftToRight)
			{
				origin = origin + (-labelSize.Width / 2 * v);
				origin = origin + labelSize.Height / 2 * vperp;
			}
			else
			{
				origin = origin + labelSize.Width / 2 * v;
				origin = origin + labelSize.Height / 2 * vperp;
			}

			//Get rotation matrix
			rotationMatrix = EdgeLabelRotationMatrix;

			//Apply the translation first
			//	rotationMatrix.Translate(-origin.x, -origin.y, MatrixOrder.Prepend);
			rotationMatrix.Translate(origin.x, origin.y, MatrixOrder.Append);

			//....may this solve the bug
			rotationMatrix.Multiply(transformBackup, MatrixOrder.Append);
			g.Transform = rotationMatrix;//.Clone();
			//rotationMatrix.Invert();
			//Now, perform the transformation
			//g.MultiplyTransform(rotationMatrix, MatrixOrder.Prepend);

			//Draw the string. The upper left corner is at (0, -sy/2)
			g.DrawString(
				label,
				font,
				brush,
				new RectangleF(new PointF(0, -labelSize.Height / 2), labelSize));

			//Restore the initial transform
			g.Transform = transformBackup;
		}

		/// <summary>
		/// Checks if pt lies within the field around the edge
		/// </summary>
		/// <param name="pt">
		/// The point checked
		/// </param>
		/// <param name="verticalDistance">
		/// The distance of the point from the straight line carrying the edge
		/// </param>
		/// <returns>
		/// true if pt is within the edge's field
		/// </returns>
		public bool WithinField(Point2F pt, out float verticalDistance)
		{
			float horizontalSignedDistance;
			Vector2F u = pt - a;
			verticalDistance = Math.Abs(Vector2F.DotProduct(u, vperp));
			horizontalSignedDistance = Vector2F.DotProduct(u, v);
			return
				Math2.Between(horizontalSignedDistance, 0, SignedRemainingDistance) &&
				verticalDistance <= Drawing.EdgeVerticalField;
		}
	}
	[Serializable]
	public class Edge
	{
		string _label; //This field is useless when _type != Labelled
		EdgeType _type;
		State _src;
		State _dst;
		EdgeGeometry _geometry;
		LabelStatus _labelStatus;
		//I think we should cache the label status so that we donot rely on COM much.
		public static Edge CreateEpsilonEdge(string charSet, State src, State dst)
		{
			return new Edge(charSet, Drawing.EpsilonLabel, EdgeType.Epsilon, src, dst);
		}
		public static Edge CreateOtherEdge(string charSet, State src, State dst)
		{
			return new Edge(charSet, Drawing.OtherLabel, EdgeType.Other, src, dst);
		}
		public static Edge CreateLabelledEdge(string charSet, string label, State src, State dst)
		{
			return new Edge(charSet, label, EdgeType.Labelled, src, dst);
		}
		private Edge(string charSet, string label, EdgeType type, State src, State dst)
		{
			string dummy;
			_label = label;
			_type = type;
			_dst = dst;
			_src = src;
			ValidateLabel(charSet, out dummy);

			if (_src != null && _dst != null)
				_geometry = new EdgeGeometry(_src, _dst);
			else
				_geometry = new EdgeGeometry();
		}
		public State Target
		{
			get { return _dst; }
			set
			{
				_dst = value;
				if (_src != null && _dst != null)
					_geometry.Reset(_src, _dst);
			}
		}
		public State Source
		{
			get { return _src; }
			set
			{
				_src = value;
				if (_src != null && _dst != null)
					_geometry.Reset(_src, _dst);
			}
		}

		public EdgeType Type
		{
			get { return _type; }
			set
			{ 
				_type = value;
				if (_type == EdgeType.Epsilon)
					_labelStatus = LabelStatus.Epsilon;
			}
		}
		public EdgeGeometry Geometry
		{
			get { return _geometry; }
		}
		public bool ChangeLabel(string charSet, string newLabel, out string errInfo)
		{
			_label = newLabel;
			return ValidateLabel(charSet, out errInfo);
		}
		public string Label
		{
			get { return _label; }
		}
        internal LabelStatus LabelStatus
        {
            get { return _labelStatus; }
        }

		/// <summary>
		/// Validates the label and determines extra information about the label such as
		/// the error string (if any) or whether the label corresponds to e-transition or not
		/// </summary>
		/// <param name="charSet">
		/// The character set of the GTG
		/// </param>
		/// <param name="status">
		/// Indicates whether the label is valid, invalid or epsilon
		/// </param>
		/// <param name="errInfo">
		/// The error string if the label is invalid
		/// </param>
		public bool ValidateLabel(string charSet, out string errInfo) //requires testing
		{
			NativeDFA dfa;
			if (_type == EdgeType.Epsilon)
			{
				_labelStatus = LabelStatus.Epsilon;
				errInfo = "";
				return true;
			}
			try
			{
				//What if _label == "". I think it should correspond to Epsilon, as well.
				dfa = new NativeDFA(_label, charSet, true);
			}
			catch (Exception ex)
			{
				_labelStatus = LabelStatus.InvalidRegex;
				errInfo = ex.Message;
				return false;
			}
			errInfo = "";
			_labelStatus = LabelStatus.ValidRegex; //or epsilon
			//Epsilon iff dfa has only the start state AND it is acceptance state
			//AND dfa all transitions from that state are to hell.

			if (dfa.TransitionTable.Length != 1) return true;
			if (!dfa.IsAcceptanceState[0]) return true;
			for (int i = 0; i < dfa.NumberOfSymbols; i++)
				if (dfa.TransitionTable[0][i] != -1) return true;

			_labelStatus = LabelStatus.Epsilon;
			return true;  
		}

		/// <summary>
		/// Determines if the label of the edge corresponds to an epsilon
		/// </summary>
		/// <param name="charSet">
		/// The character set of the GTG
		/// </param>
		/// <returns>
		/// true if the label corresponds to an epsilon
		/// </returns>
		public bool IsEpsilon() //requires testing
		{
			return _labelStatus == LabelStatus.Epsilon;
		}

		public bool IsInvalidLabel() //requires testing
		{
			return _labelStatus == LabelStatus.InvalidRegex;
		}
		public string EffectiveLabel
		{
			get
			{
				switch (_type)
				{
					case EdgeType.Labelled:
						return _label;
					case EdgeType.Epsilon:
						return Drawing.EpsilonLabel;
					case EdgeType.Other:
						return Drawing.OtherLabel;
					default:
						return null;
				}
			}
		}
		public Brush LabelBrush
		{
			get
			{
				switch (_type)
				{
					case EdgeType.Labelled:

						//Check if the label is valid or not
						if (IsInvalidLabel())
						{
							return Drawing.InvalidRegexBrush;
						}
						else
						{
							return Drawing.ValidRegexBrush;
						}
					case EdgeType.Epsilon:
						return Drawing.EpsilonRegexBrush;
						
					case EdgeType.Other:
						return Drawing.OtherRegexBrush;
					default:
						return null;
				}
			}
		}
		/// <summary>
		/// Draws the edge from the src state to Target state.
		/// </summary>
		/// <param name="g">
		/// The graphics object
		/// </param>
		/// <param name="src">
		/// The state from which this edge emanates
		/// </param>
		public void Draw(Graphics g, bool selected) //requires testing
		{
			if (_src == null || _dst == null)
				return;
			Font labelFont = null;
			Brush labelBrush = null;
			string label = "";

			//Decide label, brush, font based on edge type
			switch (_type)
			{
				case EdgeType.Labelled:

					//Check if the label is valid or not
					if (IsInvalidLabel())
					{
						labelFont = Drawing.EdgeLabelFont;
						labelBrush = Drawing.InvalidRegexBrush;
					}
					else
					{
						labelFont = Drawing.EdgeLabelFont;
						labelBrush = Drawing.ValidRegexBrush;
					}
					label = _label;

					break;

				case EdgeType.Epsilon:

					labelFont = Drawing.EdgeLabelFont;
					labelBrush = Drawing.EpsilonRegexBrush;
					label = Drawing.EpsilonLabel;

					break;
				case EdgeType.Other:

					labelFont = Drawing.EdgeLabelFont;
					labelBrush = Drawing.OtherRegexBrush;
					label = Drawing.OtherLabel;

					break;
			}
			_geometry.Draw(g, label, labelFont, labelBrush, selected);
		}
		/// <summary>
		/// If the state of an end point is changed, this method should be called
		/// to update the geometric properties of the edge
		/// </summary>
		public void StateLocationChanged()
		{
			if (_src != null && _dst != null)
				_geometry.Reset(_src, _dst);
		}
	}
	[Serializable]
	public class State
	{
		public static string DefaultName = "S";
		string _name;
		Point2F _location;
		Edge _selfEdge; //edge from the state to itself
		List<Edge> _outEdge; //Adjacency list. List of edges emanating from this state
		List<Edge> _inEdge; //List of edges ending in that state
		bool _isAccept;
		bool _isStart;

        internal int index; //I am sorry

		public static readonly GraphicsPath SelfEdgePath;
		public static readonly Vector2F SelfEdgeLabelBottomCenter; 

		static State()
		{
			SelfEdgePath = new GraphicsPath();

			Point2F origin = new Point2F(0, 0);
			//Create control points
			Point2F c1 = origin + Drawing.BV1;
			Point2F c2 = origin + Drawing.BV2;
			Point2F c3 = origin + Drawing.BV3;
			Point2F c4 = origin + Drawing.BV4;
			//Add their bezier
			SelfEdgePath.AddBezier(c1, c2, c3, c4);
			
			//The arrow tip at c4
			//The derivative of dP(t)/dt at t = 1 has the same direction as
			//the vector c4 -c3
			Vector2F bezierDir1 = (c4 - c3);
			bezierDir1.Normalize();
			Vector2F bezierNormal1 = bezierDir1.PerpendicularVector;
			
			Point2F tip1 = c4 - Drawing.ArrowHeadHorizontalComponent * bezierDir1 
								+ Drawing.ArrowHeadVerticalComponent * bezierNormal1;

			Point2F tip2 = c4 - Drawing.ArrowHeadHorizontalComponent * bezierDir1 
								- Drawing.ArrowHeadVerticalComponent * bezierNormal1;

			SelfEdgePath.AddLine(c4, tip1);
			SelfEdgePath.AddLine(c4, tip2);

			//Drawing the label
			SelfEdgeLabelBottomCenter = .5f * (Drawing.BV2 + Drawing.BV3);

		}

		public State(Point2F location)
		{
			_location = location;

			_name = DefaultName;
			_selfEdge = null;
			_outEdge = new List<Edge>();
			_inEdge = new List<Edge>();
			_isAccept = _isStart = false;
		}

		public List<Edge> OutEdges
		{
			get { return _outEdge; }
		}
		public List<Edge> InEdges
		{
			get { return _inEdge; }
		}
		public bool IsStart
		{
			get { return _isStart; }
			set { _isStart = value; }
		}

		public bool IsAccept
		{
			get { return _isAccept; }
			set { _isAccept = value; }
		}


		public Edge SelfEdge
		{
			get { return _selfEdge; }
			set { _selfEdge = value; }
		}
		public Point2F Location
		{
			get { return _location; }
			set
			{
				if (!_location.HasSameCoordinates(value))
				{
					_location = value;
					foreach (Edge edge in _inEdge)
						edge.StateLocationChanged();

					foreach (Edge edge in _outEdge)
						edge.StateLocationChanged();
				}

			}
		}
		public string Name
		{
			get { return _name; }
			set { _name = value; }
		}
		public static void DrawCircle(Graphics g, Pen pen, Point2F center, float radius)
		{
			Point2F upperLeftCorner = center.Clone();
			upperLeftCorner.Offset(-radius, -radius);

			RectangleF enclosingRect = new RectangleF(
				upperLeftCorner.x,
				upperLeftCorner.y,
				2 * radius,
				2 * radius);

			g.DrawEllipse(pen, enclosingRect);

		}
		protected void DrawSigns(Graphics g)
		{
			Point2F upperLeftCorner = _location + Drawing.PlusSignVector;
			if (IsAccept)
			{
				SizeF sz = g.MeasureString(
					Drawing.PlusString,
					Drawing.PlusMinusFont);
				g.DrawString(
					Drawing.PlusString,
					Drawing.PlusMinusFont,
					Drawing.PlusBrush,
					upperLeftCorner);
				//Update upperLeftCorner to draw '-'
				//upperLeftCorner.y += sz.Height + Drawing.MinusSignVerticalOffset;
				upperLeftCorner.x += sz.Width + Drawing.MinusSignVerticalOffset;
			}
			if (IsStart)
			{
				g.DrawString(
					Drawing.MinusString,
					Drawing.PlusMinusFont,
					Drawing.MinusBrush,
					upperLeftCorner);
			}
		}
		private void RenderStateName(Graphics g, bool selected)
		{
			StringFormat nowrap = new StringFormat(StringFormatFlags.NoWrap);
			SizeF sz = g.MeasureString(
				Name,
				Drawing.StateLabelFont,
				(int)Drawing.StateLabelWidth,
				nowrap);
			Point2F upperLeftCorner = _location.Clone();
			upperLeftCorner.Offset(-sz.Width / 2, -sz.Height / 2);

			g.DrawString(
				Name,
				Drawing.StateLabelFont,
				(selected) ? Drawing.SelectedStateLabelBrush : Drawing.StateLabelBrush,
				new RectangleF(upperLeftCorner, sz),
				nowrap);
		}
		public void DrawSelfEdge(Graphics g, string charSet, bool selfEdgeSelected)
		{
			//Render self edge
			if (_selfEdge != null)
			{
				g.TranslateTransform(_location.x, _location.y, MatrixOrder.Prepend);

				g.DrawPath(
					(selfEdgeSelected) ? Drawing.SelectedEdgePen : Drawing.EdgePen,
					SelfEdgePath);

				g.TranslateTransform(-_location.x, -_location.y, MatrixOrder.Prepend);

				//Drawing the label
				Point2F upperLeftCorner = _location + SelfEdgeLabelBottomCenter;

				StringFormat nowrap = new StringFormat(StringFormatFlags.NoWrap);
				SizeF sz = g.MeasureString(
					_selfEdge.EffectiveLabel,
					Drawing.EdgeLabelFont,
					(int)Drawing.SelfEdgeLabelMaxWidth,
					nowrap);

				upperLeftCorner.Offset(-sz.Width / 2, -sz.Height);

				g.DrawString(
					_selfEdge.EffectiveLabel,
					Drawing.EdgeLabelFont,
					_selfEdge.LabelBrush,
					new RectangleF(upperLeftCorner, sz),
					nowrap);
			}
		}
		public void DrawState(Graphics g, string charSet, bool selected)
		{
			//Draw the outer circle first
			if (selected)
				DrawCircle(g, Drawing.SelectedStateOuterCirclePen, _location, Drawing.OuterRadius);
			else
				DrawCircle(g, Drawing.OuterCirclePen, _location, Drawing.OuterRadius);

			//Draw Signs
			DrawSigns(g);

			//Draw Inner Circle if IsAccept
			if (IsAccept)
			{
				DrawCircle(g, Drawing.InnerCirclePen, _location, Drawing.InnerRadius);
			}

			//Draw incoming edge if IsStart
			if (IsStart)
			{
				Point2F incomingEndPt = _location.Clone();
				incomingEndPt.x -= Drawing.OuterRadius;
				Point2F incommingStartPt = new Point2F(
					incomingEndPt.x - Drawing.StartStateIncomingEdgeLength,
					incomingEndPt.y);
				//Arrowhead:
				Point2F topArrowHeadPt = incomingEndPt.Clone();
				Point2F bottomArrowHeadPt = incomingEndPt.Clone();
				topArrowHeadPt.Offset(-Drawing.ArrowHeadHorizontalComponent,
					Drawing.ArrowHeadVerticalComponent);
				bottomArrowHeadPt.Offset(-Drawing.ArrowHeadHorizontalComponent,
					-Drawing.ArrowHeadVerticalComponent);
				//Let's draw...
				g.DrawLine(Drawing.StartStateIncomingEdgePen,
					incommingStartPt,
					incomingEndPt);
				g.DrawLine(Drawing.StartStateIncomingEdgePen,
					incomingEndPt,
					topArrowHeadPt);
				g.DrawLine(Drawing.StartStateIncomingEdgePen,
					incomingEndPt,
					bottomArrowHeadPt);
			}

			//Draw state name
			RenderStateName(g, selected);
		}
	}

	public class EdgeMatcher
	{
		State _state;
		public EdgeMatcher(State dst)
		{
			_state = dst;
		}

		public State State
		{
			get { return _state; }
			set { _state = value; }
		}

		public bool EndsWith(Edge e)
		{
			return e.Target == _state;
		}

		public bool BeginsWith(Edge e)
		{
			return e.Source == _state;
		}
	}

	[Serializable]
	public class GTG
	{
		string _charSet;
		List<State> _states;

		public GTG()
			: this("[\\x00-\\x7F]")
		{
		}
		public GTG(string charSet)
		{
			if (ValidateCharacterSet(charSet))
			{
				_charSet = charSet;
			}
			else
				_charSet = "[\\x00-\\x7F]";
			_states = new List<State>();
		}

		/// <summary>
		/// returns the list of states maintained by the GTG. Deleting any states might
		/// cause catastrophic failures and thus prohibited and this property should
		/// not be exposed explicitly
		/// </summary>
		public List<State> States
		{
			get { return _states; }
		}

		public string CharacterSet
		{
			get { return _charSet; }
			set
			{
				if (ValidateCharacterSet(_charSet))
					_charSet = value;
			}
		}

		public void AddState(State state)
		{
			if (state != null)
				_states.Add(state);
		}

		public void RemoveState(State item)
		{
			_states.Remove(item);
			EdgeMatcher matcher = new EdgeMatcher(item);
			//Remove all edges to that state
			foreach (Edge edge in item.InEdges)
			{
				edge.Source.OutEdges.RemoveAll(matcher.EndsWith);
			}
			//Remove all edges from that state to the others
			foreach (Edge edge in item.OutEdges)
			{
				edge.Target.InEdges.RemoveAll(matcher.BeginsWith);
			}
		}
		public static bool ValidateCharacterSet(string _charSet)
		{
			try
			{
				NativeDFA dfa = new NativeDFA("[]", _charSet, false);
				return true;
			}
			catch
			{
				return false;
			}
		}

		public void Draw(Graphics g)
		{
			foreach (State state in States)
			{
				state.DrawState(g, _charSet, false);
				foreach (Edge edge in state.OutEdges)
					edge.Draw(g, false);
				state.DrawSelfEdge(g, _charSet, false);
			}
		}
		/// <summary>
		/// Searches the list of states for the states having pt inside, and returns
		/// the one with pt nearest to its center and returns the square of the distance as well.
		/// Note: If pt doesnot lie inside any state, null is returned and the distance is returned
		/// as infinity.
		/// </summary>
		/// <param name="pt"></param>
		/// <param name="nearestState"></param>
		/// <param name="radialDistanceSquared"></param>
		public bool NearestState(Point2F pt, out State nearestState, out float radialDistanceSquared)
		{
			float dist;
			radialDistanceSquared = float.PositiveInfinity;
			nearestState = null;
			foreach (State st in States)
			{
				dist = (st.Location - pt).MagnitudeSquare;
				if (dist < Drawing.OuterRadiusSquared && //Within outer circle
					dist < radialDistanceSquared) //&& nearer
				{
					radialDistanceSquared = dist;
					nearestState = st;
				}
			}
			return nearestState != null;
		}

		public bool NearestEdge(Point2F pt, out State src, out Edge nearestEdge, out float verticalDistance)
		{
			float dist;
			src = null;
			nearestEdge = null;
			verticalDistance = float.PositiveInfinity;
			foreach (State state in States)
			{
				foreach (Edge edge in state.OutEdges)
				{
					if (edge.Geometry.WithinField(pt, out dist))
						if (dist < verticalDistance)
						{
							verticalDistance = dist;
							nearestEdge = edge;
							src = state;
						}
				}
			}
			return nearestEdge != null;
		}
		public bool SelectedSelfEdge(Point2F pt, out Edge selectedSelfEdge)
		{
			Vector2F translatingVector;
			selectedSelfEdge = null;
			foreach (State state in States)
			{
				if (state.SelfEdge != null)
				{
					translatingVector = pt - state.Location;
					//Check if the translating vector touches the self edge when applied to the center
					if (State.SelfEdgePath.IsOutlineVisible(
						translatingVector.x,
						translatingVector.y,
						Drawing.SelfEdgeSelectionPen))
					{
						selectedSelfEdge = state.SelfEdge;
						return true;
					}
				}
			}
			return false;		
		}
        public bool AreAllEdgeLabelsValid()
        {
            foreach (State state in States)
            {
                foreach (Edge edge in state.OutEdges)
                {
                    if (edge.IsInvalidLabel())
                        return false;
                }
            }
            return true;
        }
        public bool RevalidateAllLabels()
        {
            bool retval = true;
            string dummy;
            foreach (State state in States)
            {
                foreach (Edge edge in state.OutEdges)
                {
                    if (!edge.ChangeLabel(_charSet, edge.Label, out dummy))
                        retval = false;
                }
            }
            return retval;
        }
        public SimplifiedGTG ToSimplifiedGTG()
        {
            if (!AreAllEdgeLabelsValid())
                return null;
            int newCount = States.Count + 2;
            SimplifiedGTG sgtg = new SimplifiedGTG();
            sgtg.States = new List<SimplifiedState>(States.Count + 2);
            sgtg.quality = new byte[newCount];
            sgtg.eliminated = new bool[newCount];
            for (int i = 0; i < newCount; i++)
            {
                sgtg.States.Add(new SimplifiedState());
                sgtg.quality[i] = 0;
                sgtg.eliminated[i] = false;
				sgtg.States[i].index = i;
            }
            for (int i = 0; i < States.Count; i++)
            {
                States[i].index = i+2;
				if (States[i].IsStart)
				{
					SimplifiedEdge edge = new SimplifiedEdge("[]", EdgeType.Epsilon,
						sgtg.States[0], sgtg.States[i + 2], LabelStatus.Epsilon);

					sgtg.States[i + 2].InEdges.Add(edge);
					sgtg.States[0].OutEdges.Add(edge);
				}
				if (States[i].IsAccept)
				{
					SimplifiedEdge edge = new SimplifiedEdge("[]", EdgeType.Epsilon,
						sgtg.States[i + 2], sgtg.States[1], LabelStatus.Epsilon);

					sgtg.States[1].InEdges.Add(edge);
					sgtg.States[i + 2].OutEdges.Add(edge);
				}
				if (States[i].SelfEdge != null)
				{
					sgtg.States[i + 2].SelfEdge = new SimplifiedEdge(
						States[i].SelfEdge.Label,
						States[i].SelfEdge.Type,
						sgtg.States[i + 2],
						sgtg.States[i + 2],
						States[i].SelfEdge.LabelStatus);
				}
            }
			foreach (State gtgState in States)
			{
				foreach (Edge gtgEdge in gtgState.OutEdges)
				{
					int srcIndex = gtgEdge.Source.index;
					int dstIndex = gtgEdge.Target.index;
					SimplifiedEdge sgtgEdge = new SimplifiedEdge(
						gtgEdge.Label, gtgEdge.Type,
						sgtg.States[srcIndex], sgtg.States[dstIndex],
						gtgEdge.LabelStatus);

					sgtg.States[srcIndex].OutEdges.Add(sgtgEdge);
					sgtg.States[dstIndex].InEdges.Add(sgtgEdge);
				}
			}
			
			return sgtg;
        }
    }

	[Serializable]
    public class SimplifiedGTG
	{
		System.Collections.Generic.List<SimplifiedState> _states;
        public byte[] quality;
        public bool[] eliminated;
		public SimplifiedGTG()
		{
            _states = null;
		}
		SimplifiedState StartState
		{
			get
			{
				return States[0];
			}
		}
		SimplifiedState FinalState
		{
			get
			{
				return States[1];
			}
		}
		/// <summary>
		/// returns the list of simplified states maintained by the SimplifiedGTG. Deleting any states might
		/// cause catastrophic failures and thus prohibited and this property should
		/// not be exposed explicitly
		/// </summary>
		public System.Collections.Generic.List<SimplifiedState> States
		{
			get { return _states; }
            set { _states = value; }
		}


		public void AddState(SimplifiedState state)
		{
			if (state != null)
				_states.Add(state);
		}

		public void RemoveState(SimplifiedState item)
		{
            /*
			_states.Remove(item);
			EdgeMatcher matcher = new EdgeMatcher(item);
			//Remove all edges to that state
			foreach (Edge edge in item.InEdges)
			{
				edge.Source.OutEdges.RemoveAll(matcher.EndsWith);
			}
			//Remove all edges from that state to the others
			foreach (Edge edge in item.OutEdges)
			{
				edge.Target.InEdges.RemoveAll(matcher.BeginsWith);
			}*/
			eliminated[item.index] = true;
		}
		public bool Process(out string regex)
		{
			bool[] visited1= new bool[States.Count];
			bool[] visited2= new bool[States.Count];
			QualifyReachableStates(States[0], visited1);
			QualifyAcceptableStates(States[1], visited2);
			if (!ValidPathExist())
			{  
				regex = "The regular language is empty";
				return false;
			}
			SimplifiedState state = FindAdjacentState();
			while (state != null)
			{
				Eliminate(state);
				state = FindAdjacentState();
			}
			//Now, the regex is the label on edge (0, 1)
			SimplifiedEdgeMatcher startToEnd = new SimplifiedEdgeMatcher(StartState);
			SimplifiedEdge regEdge = FinalState.InEdges.Find(startToEnd.BeginsWith);
			if (regEdge == null)
			{
				regex = null;
				return false;
			}
			else if (regEdge.IsEpsilon())
			{
				regex = "[]";
				return true;
			}
			else
			{
				regex = regEdge.Label;
				return true;
			}
		}

		private void Eliminate(SimplifiedState state)
		{
			SimplifiedEdge mySelfEdge = state.SelfEdge;
			SimplifiedEdge existingEdge;
			SimplifiedEdge newEdge;
			SimplifiedEdgeMatcher edgesOfMe = new SimplifiedEdgeMatcher(state);
			SimplifiedEdgeMatcher edgesOfThem = new SimplifiedEdgeMatcher(null);
			foreach (SimplifiedEdge intoMe in state.InEdges)
			{
				edgesOfThem.State = intoMe.Source;
				foreach (SimplifiedEdge outOfMe in state.OutEdges)
				{
					newEdge = SimplifiedEdge.Merge(intoMe, mySelfEdge, outOfMe);
					if (newEdge.Source == newEdge.Target) //self edge
						existingEdge = newEdge.Source.SelfEdge;
					else
						existingEdge = outOfMe.Target.InEdges.Find(edgesOfThem.BeginsWith);
					if (existingEdge != null) //Just update the existing edge and nothing new is added
					{ //Or the two labels except if both are Epsilon. In that case, no op is done.
						if (!existingEdge.IsEpsilon() || !newEdge.IsEpsilon())
						{
							existingEdge.LabelStatusValue = LabelStatus.ValidRegex;
							existingEdge.Type = EdgeType.Labelled;
							existingEdge.Label = existingEdge.Label + "|" + newEdge.Label;
						}
					}
					else//No existing edge. So, just add the new edge
					{
						//Just add the new edge
						if (newEdge.Source == newEdge.Target) //SelfEdge
							newEdge.Source.SelfEdge = newEdge;
						else
						{
							intoMe.Source.OutEdges.Add(newEdge);
							outOfMe.Target.InEdges.Add(newEdge);
						}
					}
				}
				//Remove that incoming edge
				intoMe.Source.OutEdges
					.RemoveAll(edgesOfMe.EndsWith);
			}
			//Remove all outgoing edges
			foreach (SimplifiedEdge outOfMe in state.OutEdges)
			{
				outOfMe.Target.InEdges
					.RemoveAll(edgesOfMe.BeginsWith);
			}
			//Mark that state as eliminated
			eliminated[state.index] = true;
		}
		private SimplifiedState FindAdjacentState()
		{
			foreach (SimplifiedEdge edge in StartState.OutEdges)
			{
				if (edge.Target != FinalState && quality[edge.Target.index] == 2
					&& !eliminated[edge.Target.index]) //useless since we remove all incoming edges
													  //of an eliminated state
					return edge.Target;
			}
			return null;
		}
		private bool ValidPathExist()
		{
			for (int i = 0; i < quality.Length; i++)
				if (quality[i] == 2)
					return true;
			return false;
		}
		private void QualifyReachableStates(SimplifiedState state, bool[] visited)
		{
			visited[state.index] = true;
			quality[state.index]++;
			foreach (SimplifiedEdge edge in state.OutEdges)
			{
				if (!visited[edge.Target.index])
					QualifyReachableStates(edge.Target, visited);
			}
		}

		private void QualifyAcceptableStates(SimplifiedState state, bool[] visited)
		{
			visited[state.index] = true;
			quality[state.index]++;
			foreach (SimplifiedEdge edge in state.InEdges)
			{
				if (!visited[edge.Source.index])
					QualifyAcceptableStates(edge.Source, visited);
			}
		}
	}

	[Serializable]
	public class SimplifiedState
	{
		SimplifiedEdge _selfEdge; //edge from the state to itself
		List<SimplifiedEdge> _outEdge; //Adjacency list. List of edges emanating from this state
		List<SimplifiedEdge> _inEdge; //List of edges ending in that state
		public int index;
		public SimplifiedState()
		{
			_selfEdge = null;
			_outEdge = new List<SimplifiedEdge>();
			_inEdge = new List<SimplifiedEdge>();
			
		}

		public List<SimplifiedEdge> OutEdges
		{
			get { return _outEdge; }
		}
		public List<SimplifiedEdge> InEdges
		{
			get { return _inEdge; }
		}
	

		public SimplifiedEdge SelfEdge
		{
			get { return _selfEdge; }
			set { _selfEdge = value; }
		}
	}
	public class SimplifiedEdgeMatcher
	{
		SimplifiedState _state;
		public SimplifiedEdgeMatcher(SimplifiedState state)
		{
			_state = state;
		}

		public SimplifiedState State
		{
			get { return _state; }
			set { _state = value; }
		}

		public bool EndsWith(SimplifiedEdge e)
		{
			return e.Target == _state;
		}

		public bool BeginsWith(SimplifiedEdge e)
		{
			return e.Source == _state;
		}
	}
	[Serializable]
	public class SimplifiedEdge
	{
		string _label; //This field is useless when _type != Labelled
		EdgeType _type;
		SimplifiedState _src;
		SimplifiedState _dst;
		LabelStatus _labelStatus;
		//I think we should cache the label status so that we donot rely on COM much.
		public SimplifiedEdge(string label, EdgeType type, SimplifiedState src, SimplifiedState dst, LabelStatus status)
		{
			_label = label;
			_type = type;
			_dst = dst;
			_src = src;
			_labelStatus = status;
		}
		public SimplifiedEdge()
		{
		}
		/*public SimplifiedEdge(Edge edge)
		{
			_label = edge.Label;
			_type = edge.Type;
            _src = edge.Source;
            _dst = edge.Target;
            _labelStatus = edge.LabelStatus;
		}*/
		public SimplifiedState Target
		{
			get { return _dst; }
			set
			{
				_dst = value;
			}
		}
		public SimplifiedState Source
		{
			get { return _src; }
			set
			{
				_src = value;
			}
		}
		public LabelStatus LabelStatusValue
		{
			get
			{
				return _labelStatus;
			}
			set
			{
				_labelStatus = value;
			}
		}
		public EdgeType Type
		{
			get { return _type; }
			set { _type = value; }
		}
		public string Label
		{
			get { return _label; }
            set {  _label = value; }
		}

		/// <summary>
		/// Determines if the label of the edge corresponds to an epsilon
		/// </summary>
		/// <param name="charSet">
		/// The character set of the GTG
		/// </param>
		/// <returns>
		/// true if the label corresponds to an epsilon
		/// </returns>
		public bool IsEpsilon() //requires testing
		{
			return _labelStatus == LabelStatus.Epsilon;
		}

		public bool IsInvalidLabel() //requires testing
		{
			return _labelStatus == LabelStatus.InvalidRegex;
		}
		public string LabelWithParanIfNecessary()
		{
			if (RequiresParan())
				return "(" + Label + ")";
			else return Label;
		}

		public bool RequiresParan()
		{
			char c;
			int paran = 0;
			bool delimiterFound = false;
			for (int i = 0; i < _label.Length; i++)
			{
				c = _label[i];
				switch (c)
				{
					case '|':
						if (paran <= 0)
							return true; //paren are required
						break;

					case '[':
						i++;
						delimiterFound = false;
						for (; i < _label.Length && !delimiterFound; i++)
						{
							c = _label[i];
							switch (c)
							{
								case '\\':
									i++;
									break;
								case ']':
									i--;
									delimiterFound = true;
									break;
							}
						}
						if (!delimiterFound) //end of string reached
							//Invalid regex. But, no paren are required
							return false;
						break;

					case '\\':
						i++;
						break;
					case '(':
						paran++;
						break;
					case ')':
						paran--;
						break;
					case '\"':
						i++;
						delimiterFound = false;
						for (; i < _label.Length && !delimiterFound; i++)
						{
							c = _label[i];
							switch (c)
							{
								case '\\':
									i++;
									break;
								case '\"':
									i--;
									delimiterFound = true;
									break;
							}
						}
						if (!delimiterFound) //end of string reached
							//Invalid regex. But, no paren are required
							return false;
						break;
				}
			}
			return false;
		}
		public static SimplifiedEdge Merge(SimplifiedEdge e1, SimplifiedEdge eself, SimplifiedEdge e2)
		{
			SimplifiedEdge retval;
			string mid;
			if(eself == null)
				mid = "";
			else if(eself.IsEpsilon())
			{
				mid = "";
			}
			else if(eself.Label.Length > 1)
			{
				mid = "(" + eself.Label + ")*";
			}
			else
				mid = eself.Label + "*";

			retval = new SimplifiedEdge();
			retval.Source = e1.Source;
			retval.Target = e2.Target;
//			retval.Type
			if (e1.IsEpsilon() && e2.IsEpsilon())
			{
				if (mid == "")
				{
					retval.Label = "[]";
					retval._labelStatus = LabelStatus.Epsilon;
					retval._type = EdgeType.Epsilon;
				}
				else
				{
					retval.Label = mid;
					retval._labelStatus = LabelStatus.ValidRegex;
					retval._type = EdgeType.Labelled;
				}
			}
			else if(e1.IsEpsilon())
			{
				retval.Label = mid + e2.LabelWithParanIfNecessary();
				retval._labelStatus = LabelStatus.ValidRegex;
				retval._type = EdgeType.Labelled;
			}
			else if (e2.IsEpsilon())
			{
				retval.Label = e1.LabelWithParanIfNecessary() + mid;
				retval._labelStatus = LabelStatus.ValidRegex;
				retval._type = EdgeType.Labelled;
			}
			else
			{
				retval.Label = e1.LabelWithParanIfNecessary() + mid + e2.LabelWithParanIfNecessary();
				retval._labelStatus = LabelStatus.ValidRegex;
				retval._type = EdgeType.Labelled;
			}
			return retval;
		}
	}
}