using System;

namespace Tao.OpenGl
{
	/// <summary>
	/// Summary description for Vector2F.
	/// </summary>
	[Serializable]
	public class Vector2F
	{
		public float x;
		public float y;

		public Vector2F()
		{
			x = y = 0.0f;
		}
		public Vector2F(float xx, float yy)
		{
			x = xx;
			y = yy;
		}
		public static Vector2F FromDegree(float distance, float degAngle)
		{
			float radian = (float)System.Math2.ToRadians(degAngle);
			return new Vector2F((float)(distance * System.Math.Cos(radian)),
								(float)(distance * System.Math.Sin(radian)));
		}
		public static Vector2F FromRadian(float distance, float radAngle)
		{
			return new Vector2F((float)(distance * System.Math.Cos(radAngle)),
								(float)(distance * System.Math.Sin(radAngle)));
		}
		public static Vector2F operator +(Vector2F v1, Vector2F v2)
		{
			return new Vector2F(v1.x + v2.x, v1.y + v2.y);
		}
		public static Vector2F operator -(Vector2F v1, Vector2F v2)
		{
			return new Vector2F(v1.x - v2.x, v1.y - v2.y);
		}
		public static Vector2F operator -(Vector2F v)
		{
			return new Vector2F(-v.x, -v.y);
		}
		public static Point2F operator +(Vector2F v, Point2F pt)
		{
			return new Point2F(pt.x + v.x, pt.y + v.y);
		}
		public static Point2F operator +(Point2F pt, Vector2F v)
		{
			return v + pt;
		}
		public static Point2F operator -(Point2F pt, Vector2F v)
		{
			return new Point2F(pt.x - v.x, pt.y - v.y);
		}
		public static Vector2F operator *(Vector2F v, float s)
		{
			return new Vector2F( s * v.x, s * v.y);
		}
        public static Vector2F operator *(float s, Vector2F v)
        {
            return new Vector2F(s * v.x, s * v.y);
        }
		public static bool operator ==(Vector2F v1, Vector2F v2)
		{
			bool retval = System.Math.Abs( v1.x - v2.x ) <= float.Epsilon;
			retval = retval && System.Math.Abs(v1.y - v2.y) <= float.Epsilon;
			return retval;
		}
		public static bool operator !=(Vector2F v1, Vector2F v2)
		{
			return !(v1 == v2);
		}
		public override int GetHashCode()
		{
			return ((int)x + ((int)y)<<16);
		}

		public void Scale(float s)
		{
			x = s * x;
			y = s * y;
		}
		public static float AngleBetween(ref Vector2F v1, ref Vector2F v2)
		{
			float cosZeta = DotProduct(v1, v2) / v1.Magnitude;
			cosZeta = cosZeta / v2.Magnitude;
			return (float)System.Math.Acos( cosZeta );
		}
		public override bool Equals(object obj)
		{
			Vector2F v = (Vector2F) obj;
			return this == v;
		}
		public static float DotProduct(Vector2F v1, Vector2F v2)
		{
			return v1.x*v2.x + v1.y*v2.y;
		}
		public float MagnitudeSquare
		{
			get{return x*x + y*y;}
		}
		public float Magnitude
		{
			get{return (float)System.Math.Sqrt(x*x + y*y);}
		}
		public float Angle
		{
			get{return (float)System.Math2.ToDegrees(RadianAngle);}
		}
		public float RadianAngle
		{
			get
			{return (float) System.Math.IEEERemainder(2*System.Math.PI + System.Math.Atan2(y, x), 2*System.Math.PI);}
		}
		public Vector2F NormalizedVetor
		{
			get
			{
				float m = Magnitude;
				return new Vector2F(x/m ,y/m);
			}
		}
        public Vector2F PerpendicularVector
        {
            get
            {
                return new Vector2F(-y, x);
            }
        }
        /// <summary>
        /// Returns u dotprod thisPerp
        /// </summary>
        /// <param name="u"></param>
        /// <returns></returns>
        public float PerpProduct(Vector2F u)
        {
            return u.x * (-y) + u.y * x;
        }
		public void Normalize()
		{
			float m = Magnitude;
			x = x / m;
			y = y / m;
		}
		public void Set(float xx, float yy)
		{
			x = xx;
			y = yy;
		}
		public override string ToString()
		{
			string retval = x + "i ";
			if(y<0)	retval += "- ";
			else	retval += "+ ";
			retval += System.Math.Abs(y) + "j";
			return retval;
		}


        public Vector2F Clone()
        {
            return new Vector2F(x, y);
        }
    }
}
