#include "stdafx.h"
#include "Set.h"
#include "SimpleRegexConventions.h"
#include ".\dfa.h"
#include "FAFactory.h"
#include "NFA.h"
#include "Set.h"
#include "BitSet_Min.h"
#include "LinkedList.h"
#include "Globals.h"
using namespace Automata;
#include <iostream>
#include <set>
using namespace std;
typedef LinkedList<int> StateList;
struct IntermediateState
{
	int	IndexOfHighestPriorityAcceptedNFA;
	bool ContainsCCL;
	BitSet NfaStates;
	StateList BehavioralStates;
	int *Transitions;				//externally managed.
	IntermediateState(int howManyNfaStates) : NfaStates(howManyNfaStates - 1), 
											  ContainsCCL(false),
											  IndexOfHighestPriorityAcceptedNFA(-1),
											  Transitions(NULL)
	{
	}
	BOOL IsBehaviorlyEquivalentTo(const IntermediateState *rhs) const
	{
		if(IndexOfHighestPriorityAcceptedNFA != rhs->IndexOfHighestPriorityAcceptedNFA)
			return FALSE;
		if(ContainsCCL != rhs->ContainsCCL)
			return FALSE;
		if(BehavioralStates.Count() != rhs->BehavioralStates.Count())
			return FALSE;
		StateList::const_iterator iterator = BehavioralStates.Begin();
		StateList::const_iterator endIterator = BehavioralStates.End();
		while(iterator != endIterator)
		{
			if(!rhs->NfaStates.Contains(*iterator++))
				return FALSE;
		}
		return TRUE;
	}
	~IntermediateState()
	{
	}
};
typedef LinkedList<IntermediateState *> IntermediateStateList;
struct NfaInfo
{
	NFA *Nfa;
	NfaInfo(NFA *nfa) : Nfa(nfa)
	{
	}
	int StartState() const
	{
		return Nfa->startState;
	}
	int FinalState() const
	{
		return Nfa->finalState;
	}
	BOOL IsAcceptingState(int state) const
	{
		return Nfa->finalState == state;
	}
	const NondeterministicState& GetState(int stateIndex) const
	{
		return Nfa->memory->GetState(stateIndex);
	}
	int NumberOfStates() const
	{
		return Nfa->NumberOfStates();
	}
};
void eClosure(IntermediateState *hState,
			  NFA::MemoryManager *memManager,
			  int			*nfaFinals,
			  int			nfaCount,
			  int			state)
{
	if(hState->NfaStates.Contains(state))
		return;
	hState->NfaStates.Add(state);
	const NondeterministicState &nfaState = memManager->GetState(state);
	switch( nfaState.GetType() )
	{
	/*Note: CCL Transitional states are behavioral symbol states as well. So, control must
	 *		fall to the symbol state.*/
	case NondeterministicState::CharacterClassTransition:
		hState->ContainsCCL = true;
	case NondeterministicState::SymbolTransition:
		hState->BehavioralStates.AddToBack(state);
		break;
	case NondeterministicState::EmptyTransitions:
		//Check if this is an acceptance state of a given NFA
		for(int i = 0; i<nfaCount; i++)
			if(state == nfaFinals[i])
			{
				//If it has a higher priority
				if( hState->IndexOfHighestPriorityAcceptedNFA == -1 || 
						hState->IndexOfHighestPriorityAcceptedNFA > i )
                    hState->IndexOfHighestPriorityAcceptedNFA = i;
                break;
			}
		int adj = nfaState.GetEmptyTransition1();
		if(adj != ErrorState)
			eClosure(hState, memManager, nfaFinals, nfaCount, adj);
		adj = nfaState.GetEmptyTransition2();
		if(adj != ErrorState)
			eClosure(hState, memManager, nfaFinals, nfaCount, adj);
	}	
}
DFA::DFA(void) : 	father(NULL), 
					numberOfStates(0),
					startStateIndex(0),
					transitionTable(NULL),
					actionTable(NULL)
{
	
}
void DFA::MakeRejectionMachine()
{
	if(transitionTable != NULL)
		DestroyTransitionTable();
	transitionTable = NULL;
	if(actionTable != NULL)
		DestroyActionTable();
	actionTable = NULL;
}
BOOL DFA::IsRejectionMachine() const
{
	return transitionTable == NULL;
}
DFA::DFA(	const FAFactory *creator,
			const NfaVector &nfa,
			const CodeVector &code) : father(creator),
									  transitionTable(NULL),
									  actionTable(NULL)
{
	Build(nfa, code);
}
void DFA::Reset(const FAFactory *creator,
				const NfaVector &nfa,
				const CodeVector &code)
{
	father = creator;
	Reset(nfa, code);
}
void DFA::Reset(const NfaVector &nfa,
				const CodeVector &code)
{
	if(transitionTable != NULL)
		DestroyTransitionTable();
	transitionTable = NULL;
	if(actionTable != NULL)
		DestroyActionTable();
	actionTable = NULL;
	Build(nfa, code);
}

int DFA::NextState(int state, TCHAR inputSymbol) const
{
	return transitionTable[state][father->IndexOf(inputSymbol)];
}
BOOL DFA::IsFinal(int state) const
{
	return actionTable[state] != NULL;
}
void DFA::SetAction(int state, const TCHAR *action)
{
	if( actionTable[state] != NULL )
		delete[] actionTable[state];
	actionTable[state] = CopyString(action);
}
void DFA::RemoveAction(int state)
{
	if( actionTable[state] != NULL )
		delete[] actionTable[state];
	actionTable[state] = NULL;
}
int& DFA::StartState() 
{
	return startStateIndex;
}
DFA::~DFA(void)
{
	if(transitionTable != NULL)
		DestroyTransitionTable();
	if(actionTable != NULL)
		DestroyActionTable();
}
const TCHAR* DFA::ActionOf(int state) const
{
	return actionTable[state];
}
const TCHAR* DFA::GetMatchingAction(const TCHAR *szString, int length) const
{
	
	if(IsRejectionMachine())
		return FALSE;
	int currentState = 0;//startState
	int charIndex;
	TCHAR currentChar;
	const TCHAR *end = szString + length;
	while(szString != end)
	{
		currentChar = *szString++;
		charIndex = father->IndexOf( currentChar );
		if( charIndex < 0 ) //Invalid char
			return NULL;
		currentState = transitionTable[currentState][charIndex];
		if(currentState == ErrorState)
			return FALSE;
	}
	//Now, we check currentState for acceptance...
	return actionTable[currentState];
}
const TCHAR* DFA::GetMatchingAction(const TCHAR *szString) const
{
	
	if(IsRejectionMachine())
		return FALSE;
	int currentState = 0;//startState
	int charIndex;
	TCHAR currentChar;
	while(*szString)
	{
		currentChar = *szString++;
		charIndex = father->IndexOf( currentChar );
		if( charIndex < 0 ) //Invalid char
			return NULL;
		currentState = transitionTable[currentState][charIndex];
		if(currentState == ErrorState)
			return NULL;
	}
	//Now, we check currentState for acceptance...
	return actionTable[currentState];
}
BOOL DFA::Match(const TCHAR *szString, int length) const
{
	return GetMatchingAction(szString, length) != NULL;
}
BOOL DFA::Match(const TCHAR *szString) const
{
	return GetMatchingAction(szString) != NULL;
}
void DFA::DestroyTransitionTable()
{
	for(int i=0;i<numberOfStates;i++)
		if(transitionTable[i]!=NULL)
		{
			delete[] transitionTable[i];
			transitionTable[i] = NULL;
		}
	delete[] transitionTable;
}
void DFA::DestroyActionTable()
{
	for(int i=0;i<numberOfStates;i++)
		if(actionTable[i]!=NULL)
		{
			delete[] actionTable[i];
			actionTable[i] = NULL;
		}
	delete[] actionTable;
}
void DFA::Build(const NfaVector &nfa,
				const CodeVector &code)
{
	int numberOfSymbols = father->NumberOfSymbols();
	int nfaStatesCount = 0;
	int nfaCount = nfa.size();
	int *finalState = new int[nfaCount];
	if(nfaCount < 1)
	{
		MakeRejectionMachine();
		return;
	}
	for(int j=0; j<nfaCount; j++)
	{
		nfaStatesCount += nfa[j]->NumberOfStates();
		finalState[j] = nfa[j]->finalState;
	}
	NFA::MemoryManager *memManager = nfa[0]->memory;
	//Each IntermediateState represent a DFA state defining the 
	//corresponding set of NFA states.
	IntermediateState *tempState = new IntermediateState(nfaStatesCount);
	IntermediateStateList states;	//Linked list of DFA States
	IntermediateState *startState = 
					new IntermediateState(nfaStatesCount);
	//The initial state includes all the initial states of all nfas
	for(int j=0; j<nfaCount; j++)
	{
		eClosure(startState,  memManager, finalState, nfaCount, nfa[j]->startState);
	}
	states.AddToBack(startState);
    
	IntermediateStateList::iterator	i = states.Begin();
	IntermediateState *currentState = NULL;
	//Determine the behaviour(transitions) of DFA states
	//and add newly recognized DFA states to the end of the list
	//for later processing.
	while(i != states.End())
	{
		currentState = *i;
		currentState->Transitions = new int[numberOfSymbols];
		memset(currentState->Transitions, ErrorState, numberOfSymbols*sizeof(int));
		if(	currentState->ContainsCCL || 
			currentState->BehavioralStates.Count() >= numberOfSymbols)
		{
//			FAFactory::const_iterator afterEndIterator = father->End();
//			FAFactory::const_iterator symbolIterator = father->Begin();
			TCHAR currentSymbol;
			int currentSymbolIndex = 0;
			while(currentSymbolIndex < numberOfSymbols)
			{
 				currentSymbol = father->CharacterOf(currentSymbolIndex);
				StateList::iterator bstateIterator = currentState->BehavioralStates.Begin();
				StateList::iterator bstateEnd = currentState->BehavioralStates.End();
				while(bstateIterator != bstateEnd)
				{
					const NondeterministicState &currentBState = 
						memManager->GetState(*bstateIterator++);
					int nextStateIndex = currentBState.NextState(father, currentSymbol);
					if(nextStateIndex != ErrorState)
						eClosure(tempState,  memManager, finalState, nfaCount, nextStateIndex);
				}
				int nextIStateIndex;
				//if the next state is the ErrorState(if it is nonaccepting and have no outside 
				//transitions, then set state(currentSymbol) = ErrorState
				if(tempState->IndexOfHighestPriorityAcceptedNFA == -1 && tempState->BehavioralStates.Count() == 0)
				{
					nextIStateIndex = ErrorState;
					tempState->NfaStates.SetToEmptySet();
				}
				else
				{
					//if there is already a state that is behaviorly equivalent to tempState, 
					//then no need to create a new state. Just set it.
					IntermediateStateList::iterator iterIStates = states.Begin();
					IntermediateStateList::iterator endIStates = states.End();
					nextIStateIndex = 0;
					BOOL alreadyExists = FALSE;
					while(iterIStates != endIStates)
					{
						if(tempState->IsBehaviorlyEquivalentTo(*iterIStates++))
						{
							alreadyExists = TRUE;
							iterIStates = endIStates; //end the loop
						}
						else
						{
							nextIStateIndex++;
						}
					}
					if(!alreadyExists)
					{
						tempState->Transitions = new int[numberOfSymbols];
						states.AddToBack(tempState);
						tempState = new IntermediateState(nfaStatesCount);
					}
					else //if alreadyExists
					{
						//either delete it or reset it
						delete tempState;
						tempState = new IntermediateState(nfaStatesCount);						
					}
				}
				currentState->Transitions[currentSymbolIndex] = nextIStateIndex;
				currentSymbolIndex++;
			}// end of for each symbol in E
		}//end of if |BStats| > |E|
		else
		{
			TCHAR currentSymbol;
			int currentSymbolIndex;
			//Iterate through the behavioural symbolic states and determine 
			//their transitions.
			/*
				for each symbolic NFA state in the current DFA state in the list
					- if it has been checked before, skip it to the next symbolic
						state
					- else we are sure that there is no previous symbolic state
						with the same symbol. If there were, then the current
						state would have been visited and had its status as
						checked.
						IntermediateState = {} //The DFA state we shall move to
						So, from the current symbolic state, we go through the 
						remaining symbolic states and do the following:
							If it has a different symbol, skip it.
							If it has the same symbol, 
								set it as checked.
								Determine its target NFA state s
								call eClosure( s, IntermediateState )
						If the IntermediateState has been seen before,
							set transiton[ IndexOf(Symbol) ] = Index of that
							Intermediate State,
						else
							Append that IntermediateState to the end of the
							DFA state list, and set transition[ IndexOf(Symbol) ]
								= Index of that Intermediate State
			*/
			StateList::iterator bstateIterator = currentState->BehavioralStates.Begin();
			StateList::iterator bstateEnd = currentState->BehavioralStates.End();
			while(bstateIterator != bstateEnd)
			{
				const NondeterministicState &currentBState = 
												memManager->GetState(*bstateIterator);
				currentSymbol = currentBState.GetTransitionSymbol();
				currentSymbolIndex = father->IndexOf(currentSymbol);
				//StateList::iterator bsIterator = bstateIterator++;
				bstateIterator++;
				StateList::iterator bsIterator = currentState->BehavioralStates.Begin();
				while(bsIterator != bstateEnd)
				{
					const NondeterministicState &cbState = 
						memManager->GetState(*bsIterator++);
					int nextStateIndex = cbState.NextState(father, currentSymbol);
					if(nextStateIndex != ErrorState)
						eClosure(tempState, memManager, finalState, nfaCount, nextStateIndex);
				}
				int nextIStateIndex;
				//if the next state is the ErrorState(if it is nonaccepting and have no outside 
				//transitions, then set state(currentSymbol) = ErrorState
				if(tempState->IndexOfHighestPriorityAcceptedNFA == -1 && tempState->BehavioralStates.Count() == 0)
				{
					nextIStateIndex = ErrorState;
					tempState->NfaStates.SetToEmptySet();
				}
				else
				{
					//if there is already a state that is behaviorly equivalent to tempState, 
					//then no need to create a new state. Just set it.
					IntermediateStateList::iterator iterIStates = states.Begin();
					IntermediateStateList::iterator endIStates = states.End();
					nextIStateIndex = 0;
					BOOL alreadyExists = FALSE;
					while(iterIStates != endIStates)
					{
						if(tempState->IsBehaviorlyEquivalentTo(*iterIStates++))
						{
							alreadyExists = TRUE;
							iterIStates = endIStates; //end the loop
						}
						else
						{
							nextIStateIndex++;
						}
					}
					if(!alreadyExists)
					{
						states.AddToBack(tempState);
						tempState = new IntermediateState(nfaStatesCount);
					}
					else //if alreadyExists
					{
						//either delete it or reset it
						delete tempState;
						tempState = new IntermediateState(nfaStatesCount);						
					}
				}
				currentState->Transitions[currentSymbolIndex] = nextIStateIndex;
			}// end of for each symbol in E
		}//end of if( |BStates| < |E| and no character classes in this state )
		i++;
    }
	delete tempState;
	//Now, we must construct the transition table for the dfa
	if(transitionTable != NULL)
		DestroyTransitionTable();
	transitionTable = NULL;
	if(actionTable != NULL)
		DestroyActionTable();
	actionTable = NULL;
	this->numberOfStates = states.Count();
	transitionTable = new int*[this->numberOfStates];
	actionTable = new TCHAR*[this->numberOfStates];
	IntermediateStateList::iterator iter = states.Begin();
	for(int j = 0; j < this->numberOfStates; j++, iter++)
	{
		transitionTable[j] = (*iter)->Transitions;
		if((*iter)->IndexOfHighestPriorityAcceptedNFA >= 0)
		{
			TCHAR *action = code[ (*iter)->IndexOfHighestPriorityAcceptedNFA ];
			actionTable[j] = CopyString(action);						
		}
		else
			actionTable[j] = NULL;
		delete *iter;
	}
//	cout<<endl<<_T("END OF DESTRUCTION")<<endl;
	startStateIndex = 0;
}
void DFA::Minimize()
{
	InitializeGroups();
	int Last_numGrroups; //last number of groups
	int Curr_Ch;//current character
	Set **current_Group;//BitSet_Min **current_Group;
	Set **NewPartition; //new Partition being created
	int first_state; 
	int next_state;
	int Target_of_first; 
	int Target_of_second;

	do
	{
		Last_numGrroups=Numgroups;
		for(current_Group=&Groups[0];current_Group<&Groups[Numgroups];++current_Group)
		{
			if(((BitSet_Min*)*current_Group)->Unity_Count())//if(num_ele(*current_Group)<=1) //unknown function
				continue;
			NewPartition = &Groups[Numgroups];
			*NewPartition = new BitSet_Min(numberOfStates);
			//next_member(NULL);//unknown nextmember
			first_state = ((BitSet_Min*)*current_Group)->NextMemberFrom(0);//nextmember Not known
			//int Set_iterator=first+1;
			next_state=first_state;
			while((next_state=((BitSet_Min*)*current_Group)->NextMemberFrom(next_state+1))>=0)
			{
				
				//for(Curr_Ch=this->father->startSymbol;Curr_Ch<this->father->startSymbol+Curr_Ch<this->father->NumberOfSymbols();Curr_Ch++) //MAXCHARS unknown
				for(Curr_Ch=0;Curr_Ch<this->father->NumberOfSymbols();Curr_Ch++) 
				{
					Target_of_first = transitionTable[first_state][Curr_Ch];
					Target_of_second = transitionTable[next_state][Curr_Ch];
					if(Target_of_first!=Target_of_second&&(Target_of_first==ErrorState||Target_of_second==ErrorState||InverseGroups[Target_of_first]!=InverseGroups[Target_of_second]))
					{
						//*NewPartition = new BitSet_Min(numberOfStates);//mostagadda
						(*current_Group)->Remove(next_state);
						(*NewPartition)->Add(next_state);
						InverseGroups[next_state]=Numgroups;
						break;
					}

				}
			}
			if((*NewPartition)->IsEmpty())
			{
				//delete newpartition set
				//delete NewPartition;
			}
			else
			{
				++Numgroups;
			}
		}
	}while(Last_numGrroups!=Numgroups);
	if(Numgroups<NumberOfStates())
		RecontructTransitionTable();
	
}
//Helper functions for miunimization

void DFA::InitializeGroups()
{
	Groups = new Set*[numberOfStates];
	InverseGroups = new int [numberOfStates];
	Automata::Set **last;
	int i,j;
	last = &Groups[0];
	Numgroups=0;
	bool Group_Exists;
	for(i=0;i<numberOfStates;i++) 
	{
		Group_Exists=false;
		for(j=i;--j>=0;)
		{
			if(custom_strcmp(*(actionTable+i),*(actionTable+j))==0)//Variables undefined
			{
				Groups[InverseGroups[j]]->Add(i);
				InverseGroups[i]=InverseGroups[j];
				Group_Exists=true;
				break;
			}
		}

		if(!Group_Exists)//means create a new group
		{
			*last=new BitSet_Min(numberOfStates);
			(*last)->Add(i);  //Not working
			InverseGroups[i] = Numgroups++;
			++last;

		}
	}
	//for Error Tracking purposes
	/*for(int i=0;i<Numgroups;i++)
	{
		cout<<"Group "<<i<<endl<<"============"<<endl;
		BitSet_Min* Gi=(BitSet_Min*)Groups[i];
		int state=Gi->NextMemberFrom(0);
		cout<<state<<" ";
		while((state=Gi->NextMemberFrom(state+1))!=-1)
		{
			cout<<state<<" "		;	
		}
		cout<<endl;

		
	}*/
}
void DFA::PrintTransitionTable()
{
	/*FAFactory charSet1('0', '2');
	father =&charSet1;*/
	//cout<<"It's me the transistion table\n";
	int i;
	for(i=0;i< this->father->NumberOfSymbols()*16;i++)
	{
		cout<<'-';
	}
	cout<<endl<<"|\t|";
	for(i=0;i< this->father->NumberOfSymbols();i++)
	{
		cout<<"   "<<i<<"\t";
	}
	cout<<"|\tCode\t|"<<endl;
	for(i=0;i< this->father->NumberOfSymbols()*16;i++)
	{
		cout<<'-';
	}
	cout<<endl;
	for(i=0;i< this->numberOfStates;i++)
	{
		cout<<"|  "<<i<<"\t|";
		for(int j=0;j< this->father->NumberOfSymbols();j++)
		{
			cout<<"   "<<transitionTable[i][j]<<"\t";
		}
		cout<<"|";
		if(IsFinal(i))
		{
			cout<<"\t"<<actionTable[i]<<"\t|";
		}
		else
			cout<<"\t   \t|";
		cout<<endl;
		
	}
	for(i=0;i< this->father->NumberOfSymbols()*16;i++)
	{
		cout<<'-';
	}
	cout<<endl;
}
int DFA::custom_strcmp(TCHAR *s1, TCHAR *s2)
{
	if(s1==0&&s2==0)
		return 0;
	else
		if((s1==0&&s2!=0) || (s2==0&&s1!=0))
			return -1;
 	while (*s1 == *s2++) 
	{
		if (*s1++ == (TCHAR)0) 
		{
			return (0);
		}
	}
	return (((unsigned long)*s1) - ((unsigned long)*(--s2)));
}
void DFA::RecontructTransitionTable()
{
	Automata::Set **current;
	int state=0;
	int i;
	int *Source,*Destination;
	int **NewTransisitionTable = new int*[Numgroups];
	TCHAR **NewactionTable = new TCHAR*[Numgroups];
	for(current=&Groups[Numgroups];--current>=Groups;)
	{
		//Destination=&NewTransisitionTable[current-Groups][0];
		//Destination=NewTransisitionTable[current-Groups];
		//Destination=new int [father->NumberOfSymbols()];
		NewTransisitionTable[current-Groups]=new int [father->NumberOfSymbols()];
		Destination=NewTransisitionTable[current-Groups];
		state = ((BitSet_Min*)(*current))->NextMemberFrom(0);
		Source = &transitionTable[state][0];
		NewactionTable[current-Groups] = actionTable[state];

		for(i=father->NumberOfSymbols();--i>=0;Source++,Destination++)
		{
			*Destination=(*Source==ErrorState)? ErrorState:InverseGroups[*Source];
		}
	}
	DestroyTransitionTable();
	//DestroyActionTable();
	delete[] actionTable;
	delete[] InverseGroups;
	for(int i=0;i<Numgroups;i++)
		delete Groups[i];//Groups[i]: pointer to the ith set.
	delete [] Groups;
	transitionTable=NewTransisitionTable;
	actionTable=NewactionTable;
	this->numberOfStates=Numgroups;
    



}
void DFA::AssignTransTable()
{
	father =new FAFactory('0','2');
	this->numberOfStates=8;
	transitionTable = new int* [8];
	int* arr0=new int[3];
	arr0[0]=1;arr0[1]=2;arr0[2]=-1;
  	transitionTable[0] = arr0;
	int* arr1=new int[3];
	arr1[0]=4;arr1[1]=5;arr1[2]=-1;
	transitionTable[1] = arr1;
	int* arr2=new int[3];
	arr2[0]=3;arr2[1]=-1;arr2[2]=-1;
	transitionTable[2] = arr2;
	int* arr3=new int[3];
	arr3[0]=-1;arr3[1]=-1;arr3[2]=-1;
	transitionTable[3] = arr3;
	int* arr4=new int[3];
	arr4[0]=4;arr4[1]=2;arr4[2]=-1;
	transitionTable[4] = arr4;
	int* arr5=new int[3];
	arr5[0]=6;arr5[1]=-1;arr5[2]=-1;
	transitionTable[5] = arr5;
	int* arr6=new int[3];
	arr6[0]=7;arr6[1]=-1;arr6[2]=-1;
	transitionTable[6] = arr6;
	int* arr7=new int[3];
	arr7[0]=7;arr7[1]=-1;arr7[2]=-1;
	transitionTable[7] = arr7;

	actionTable= new TCHAR* [8];
	*(actionTable)=*(actionTable+1)=*(actionTable+2)=*(actionTable+4)=0;
	*(actionTable+3)=_T("YES");
	*(actionTable+5)=_T("YES");
	*(actionTable+6)=_T("YES");
	*(actionTable+7)=_T("YES");
}


/**************************************************************************************/
//COMPRESSION

CompressedDFA* DFA::Compress()
{
	//A case statement that, according to the user's preferences, selects the suitable compression technique.
	//Each compression technique is a private function called from here.
	//The private function will assign the resulting table to the suitable compressed DFA class.

	switch( DFA::CompressionTechnique )
	{
		case DFA::DFACompressionTechnique::MostCompact:
			return MostCompactTechnique( 0 );    //Dummy, temporarily
		case DFA::DFACompressionTechnique::Pairs_Normal:
			return CompressToPairs_Normal( 0 );  //Dummy, temporarily
		case DFA::DFACompressionTechnique::Pairs_Default:
			return CompressToPairs_Default( 0 ); //Dummy, temporarily
		case DFA::DFACompressionTechnique::RedundancyRemoved:
			return CompressToRedundancyRemoved();
		case DFA::DFACompressionTechnique::Aho:
			return CompressToAho();
	}
}

CompressedDFA_Pairs_Normal* DFA::CompressToPairs_Normal( int threshold )
{
	CompressedDFA_Pairs_Normal *compressedDFA = new CompressedDFA_Pairs_Normal( this );
	
	int row, column, numberOfTransitions;
	int currentTransitionTableCell;

	int rowsCount = NumberOfStates();
	int columnsCount = father->NumberOfSymbols();

	compressedDFA->TransitionListVector = new int*[rowsCount];
	//the memory size is a variable that holds the number of bytes used by the compressed DFA
	//initially the variable equals the number of states ..
	//increment the variable every time you add a cell, then multiply by the size of int at the end
	compressedDFA->memorySize = numberOfStates;

	for( row = 0; row < rowsCount; row++ )
	{
		numberOfTransitions = 0;
		currentTransitionTableCell = transitionTable[row][0];
		for( column = 0;
			 column < columnsCount;
			 currentTransitionTableCell = transitionTable[row][++column] )
			     if( currentTransitionTableCell != -1 )
					 numberOfTransitions++;

		if( numberOfTransitions > 0 ) //If not a hell state
		{
			if( numberOfTransitions > threshold ) //If no compression is to be done
			{
				compressedDFA->memorySize += columnsCount + 1;

				compressedDFA->TransitionListVector[row] = new int[columnsCount + 1];
				compressedDFA->TransitionListVector[row][0] = 0;

				for( int cell=1; cell<=columnsCount; cell++ )
					compressedDFA->TransitionListVector[row][cell] = transitionTable[row][cell-1];
			}
			else //Let's compress
			{
				compressedDFA->memorySize += numberOfTransitions*2 + 1;

				compressedDFA->TransitionListVector[row] = new int[numberOfTransitions*2 + 1];
				compressedDFA->TransitionListVector[row][0] = numberOfTransitions;

				int currentPair = 0;
				for( int cell=0; cell<columnsCount; cell++ )
					if( transitionTable[row][cell] != -1 )
					{
						compressedDFA->TransitionListVector[row][currentPair*2 + 1] = cell;
						compressedDFA->TransitionListVector[row][currentPair*2 + 2] = transitionTable[row][cell];
						currentPair++;
					}    
			}
		}
		else
			compressedDFA->TransitionListVector[row] = NULL;
	}
	return compressedDFA;
}
CompressedDFA_Pairs_Default* DFA::CompressToPairs_Default( int threshold )
{
	CompressedDFA_Pairs_Default *compressedDFA = new CompressedDFA_Pairs_Default( this );

	int row, column, numberOfTransitions, currentTransitionTableCell, *stateCounter, defaultState, maximumOccurrence;

	int transitionListVectorSize = 0;

	int rowsCount = NumberOfStates();
	int columnsCount = father->NumberOfSymbols();

	compressedDFA->TransitionListVector = new int*[rowsCount];

	for( row = 0; row < rowsCount; row++ )
	{
		numberOfTransitions = 0;
		currentTransitionTableCell = transitionTable[row][0];
		for( column = 0;
			 column < columnsCount;
			 currentTransitionTableCell = transitionTable[row][++column] )
			     if( currentTransitionTableCell != -1 )
				     numberOfTransitions++;

		if( numberOfTransitions > 0 ) //If not a hell state
		{
			if( numberOfTransitions > threshold ) //If no compression is to be done
			{
				//transitionListVectorSize += columnsCount + 1;

				compressedDFA->TransitionListVector[row] = new int[columnsCount+1];
				compressedDFA->TransitionListVector[row][0] = 0;

				for( int cell=1; cell<=columnsCount; cell++ )
					compressedDFA->TransitionListVector[row][cell] = transitionTable[row][cell-1];
			}
			else //Let's compress
			{
				//transitionListVectorSize += numberOfTransitions*2 + 2;

				compressedDFA->TransitionListVector[row] = new int[numberOfTransitions*2 + 2];
				compressedDFA->TransitionListVector[row][0] = numberOfTransitions;

				int currentPair = 0;
				for( int cell = 0; cell<columnsCount; cell++ )
					if( transitionTable[row][cell] != -1 )
					{
						compressedDFA->TransitionListVector[row][currentPair*2 + 2] = cell;
						compressedDFA->TransitionListVector[row][currentPair*2 + 3] = transitionTable[row][cell];
						currentPair++;
					};

				//Determining the default state
				stateCounter = new int[numberOfStates];
				memset( stateCounter, 0, sizeof(stateCounter) );

				for( int cellIndex = 2; cellIndex <= currentPair*2 + 1; cellIndex += 2 )
					stateCounter[ compressedDFA->TransitionListVector[row][cellIndex] ]++;

				defaultState = -1;
				maximumOccurrence = -1;
				for( int state = 0; state < numberOfStates; state++ )
					if( maximumOccurrence < stateCounter[state] )
					{
						maximumOccurrence = stateCounter[state];
						defaultState = state;
					};

				compressedDFA->TransitionListVector[row][1] = defaultState;
			}
		}
		else
			compressedDFA->TransitionListVector[row] = NULL;
	}

	return compressedDFA;
}

CompressedDFA_RedundancyRemoved* DFA::CompressToRedundancyRemoved()
{
	int i,j,k;
	CompressedDFA_RedundancyRemoved *compressedDFA 
		= new CompressedDFA_RedundancyRemoved( this );

	int originalColumnsCount = father->NumberOfSymbols();
	int originalRowsCount = numberOfStates;
	int reducedColumnsCount = 0, reducedRowsCount = 0,
		currentColumn, currentlyInspectedColumn, currentRow, currentlyInspectedRow,
		nextColumn = 0, nextRow = 0;

	//allocate the necessary memory for compression
	compressedDFA->ColumnMap = new int[originalColumnsCount];
	for(i=0;i<originalColumnsCount;i++)
		compressedDFA->ColumnMap[i] = -1;

	compressedDFA->RowMap = new int[originalRowsCount];
	for(i=0;i<originalRowsCount;i++)
		compressedDFA->RowMap[i] = -1;

	compressedDFA->MappedTransitionTable = new int*[originalRowsCount];
	for(i=0;i<originalRowsCount;i++)
	{
		compressedDFA->MappedTransitionTable[i] = new int[originalColumnsCount];
		for(j=0;j<originalColumnsCount;j++)
			compressedDFA->MappedTransitionTable[i][j] = -1;
	}
	
	//compress rows
	for(i=0;i<originalRowsCount;i++)
	{
		//now we are pointing to the row number i..
		//search in the reduced row set for a row similar to the row i
		for(j=0;j<reducedRowsCount;j++)
		{
			//compare the two rows
			for(k=0;k<originalColumnsCount;k++)
				if(transitionTable[i][k] 
				!= compressedDFA->MappedTransitionTable[j][k])
					break;

			//are they equal?
			if(k == originalColumnsCount)//if yes, break
				break;
		}

		//are they equal?
		if(j != reducedRowsCount)//yes, the row i = the reduced row j
		{
			//let the ith row map item point to the reduced row j
			compressedDFA->RowMap[i] = j;
		}
		else//no, this is a new row
		{
			//add the ith row to the reduced row set
			for(k=0;k<originalColumnsCount;k++)
				compressedDFA->MappedTransitionTable[reducedRowsCount][k] = 
				transitionTable[i][k];
			//let the ith item in the row map point to the new row
			compressedDFA->RowMap[i] = reducedRowsCount;
			//increment the count of the reduced rows set
			reducedRowsCount++;
		}
	}

	//compress columns
	for(i=0;i<originalColumnsCount;i++)
	{
		//now we are pointing to the column number i..
		//search in the reduced column set for a column similar to the column i
		for(j=0;j<reducedColumnsCount;j++)
		{
			//compare the two rows
			for(k=0;k<reducedRowsCount;k++)
				if(compressedDFA->MappedTransitionTable[k][i] 
				!= compressedDFA->MappedTransitionTable[k][j])
					break;

			//are they equal?
			if(k == reducedRowsCount)//if yes, break
				break;
		}

		//are they equal?
		if(j != reducedColumnsCount)//yes, the row i = the reduced row j
		{
			//let the ith row map item point to the reduced row j
			compressedDFA->ColumnMap[i] = j;
		}
		else//no, this is a new row
		{
			//add the ith row to the reduced row set
			for(k=0;k<reducedRowsCount;k++)
				compressedDFA->MappedTransitionTable[k][reducedColumnsCount] = 
				compressedDFA->MappedTransitionTable[k][i];
			//let the ith item in the row map point to the new row
			compressedDFA->ColumnMap[i] = reducedColumnsCount;
			//increment the count of the reduced rows set
			reducedColumnsCount++;
		}
	}

	compressedDFA->ReducedColumnsCount = reducedColumnsCount;
	compressedDFA->ReducedRowsCount    = reducedRowsCount;

	return compressedDFA;
}
int DFA::ColumnsAreEquivalent(int baseColumn, int inspectedColumn, int originalRowsCount)
{
	int currentRow;
	for( currentRow = 0; currentRow < originalRowsCount; currentRow++ )
		if( transitionTable[currentRow][baseColumn] != transitionTable[currentRow][inspectedColumn] )
			break;
	return ( currentRow == originalRowsCount );
}
int DFA::RowsAreEquivalent(int baseRow, int inspectedRow, int originalColumnsCount)
{
	int currentColumn;
	for( currentColumn = 0; currentColumn < originalColumnsCount; currentColumn++ )
		if( transitionTable[baseRow][currentColumn] != transitionTable[inspectedRow][currentColumn] )
			break;
	return ( currentColumn == originalColumnsCount );
}

CompressedDFA_Aho* DFA::CompressToAho()
{
	CompressedDFA_Aho *compressedDFA = new CompressedDFA_Aho( this );

	return compressedDFA;
}

CompressedDFA* DFA::MostCompactTechnique( int threshold )
{
	int compressToPairs_Default;
	int compressToPairs_Normal;
    int compressToRedundancyRemoved;

	//Pairs
	int row, column, numberOfTransitions;
	int currentTransitionTableCell;

	int rowsCount = NumberOfStates();
	int columnsCount = father->NumberOfSymbols();

	for( row = 0; row < rowsCount; row++ )
	{
		numberOfTransitions = 0;
		currentTransitionTableCell = transitionTable[row][0];
		for( column = 0;
			 column < columnsCount;
			 currentTransitionTableCell = transitionTable[row][++column] )
			     if( currentTransitionTableCell != -1 )
					 numberOfTransitions++;

		if( numberOfTransitions > 0 ) //If not a hell state
			if( numberOfTransitions > threshold ) //If no compression is to be done
				compressToPairs_Normal += columnsCount + 1;
			else //Let's compress
				compressToPairs_Normal += numberOfTransitions*2 + 1;
	}

	compressToPairs_Normal += numberOfStates;                          //To include the TransitionListVector itself
	compressToPairs_Default = compressToPairs_Normal + numberOfStates; //The only difference is the second extra column in all rows

	//Redundancy Removed
	CompressedDFA_RedundancyRemoved* tempCompressedDFA = CompressToRedundancyRemoved();
	compressToRedundancyRemoved = (tempCompressedDFA->ReducedRowsCount) * (tempCompressedDFA->ReducedColumnsCount)
								  + rowsCount + columnsCount;
	if( compressToPairs_Normal < compressToRedundancyRemoved )
		return CompressToPairs_Normal( 0 ); //Threshold	//tempCompressedDFA is a memory leak, I think.
	else
		return tempCompressedDFA;
}