#pragma once
#include <vector>
#include "stdafx.h"
#include "set.h"
#include "compressedDFA.h"

using namespace std;
using namespace Automata;

namespace Automata
{
	class FAFactory;
	class NFA;
	class CompressedDFA;
	class CompressedDFA_Aho;
	class CompressedDFA_Pairs_Default;
	class CompressedDFA_Pairs_Normal;
	class CompressedDFA_RedundancyRemoved;
class DFA
	{
	friend class FAFactory;
	friend class CompressedDFA_Aho;
	friend class CompressedDFA_Pairs_Default;
	friend class CompressedDFA_Pairs_Normal;
	friend class CompressedDFA_RedundancyRemoved;
	private:
		const FAFactory		*father;
		int					numberOfStates;
		int					startStateIndex;//Always zero
		TCHAR*				*actionTable;	//a state qi is final if actionTable[i] != NULL.
		int					**transitionTable;   // transitionTable[qi][cr] = qj

		/***************************************/
		/*Helper Methods*/
		void	Build(const NfaVector &nfa,
				const CodeVector &code);

		/****************************************************/ 
		/****** DFA minimization variables & functions ******/
		//data
		Set **Groups; //Array of sets,each set contains states
		int Numgroups;  // number of groups...
		int* InverseGroups; //index is the state and returns its group
		//helperfunctions
		void InitializeGroups();
		int custom_strcmp(TCHAR *s1, TCHAR *s2);
		void RecontructTransitionTable();

		/***************************************/
		/***************************************/
	public:
		DFA();
		DFA(const FAFactory *creator,
			const NfaVector &nfa,
			const CodeVector &code);

		int		NextState(int state, TCHAR inputSymbol) const;
		BOOL	IsFinal(int state) const;
		void	SetAction(int state, const TCHAR *action);
		void	RemoveAction(int state);
		void	ClearAllFinals();
		int		StartState() const;
		int&	StartState();
		void	Reset(const NfaVector &nfa,
				const CodeVector &code);
		void	Reset(	const FAFactory *creator,
						const NfaVector &nfa,
						const CodeVector &code);
		void	MakeRejectionMachine();
		BOOL	IsRejectionMachine() const;
		BOOL	Match(const TCHAR *szString, int length) const;
		BOOL	Match(const TCHAR *szString) const;
		const TCHAR* GetMatchingAction(const TCHAR *szString, int length) const;
		const TCHAR* GetMatchingAction(const TCHAR *szString) const;
		const TCHAR* ActionOf(int state) const;
		void	DestroyTransitionTable();
		void	DestroyActionTable();

		//BEGIN accessors to be used by compressed DFA's
		const FAFactory*  Father()
		{
			return father;
		}
		int	NumberOfStates()const 
		{
			return numberOfStates;
		}
		int NumberOfSymbols()const
		{
			return father->NumberOfSymbols();
		}
		int NumberOfBytes() const
		{
			return NumberOfStates() * NumberOfSymbols() * sizeof(int);
		}
		int	StartStateIndex()const
		{
			return startStateIndex;
		}
		TCHAR* *ActionTable()const
		{
			return actionTable;
		}
		const int * const* TransitionTable() const
		{
			return transitionTable;
		}
		//END accessors

		~DFA(void);
		//for minmization
		void Minimize();
		void PrintTransitionTable();
		void AssignTransTable();

		//Code Generation
		void GenerateCodeCpp(
			tofstream& tout,
			tofstream& tout_h,
			Dictionary& options,
			const tstring& HeaderTopSection,
			const tstring& ClassDefinition,
			const tstring& CppTopSection,
			const tstring& CppBottomSection
			);
		void GenerateCodeCs(
			tofstream& tout,
			Dictionary options,
			const tstring& HeaderTopSection,
			const tstring& ClassDefinition,
			const tstring& CppTopSection,
			const tstring& CppBottomSection
			);
		void GenerateCodeJava(
			tofstream& tout,
			Dictionary options,
			const tstring& HeaderTopSection,
			const tstring& ClassDefinition,
			const tstring& CppTopSection,
			const tstring& CppBottomSection
			);

		//Compression		
		int  CompressionTechnique;
		enum DFACompressionTechnique
		{
			MostCompact,
			NoCompression,
			RedundancyRemoved,
			Pairs_Normal,
			Pairs_Default,
			Aho
		};

		CompressedDFA* Compress();
		CompressedDFA_Pairs_Normal* CompressToPairs_Normal
			( int threshold );
		CompressedDFA_Pairs_Default* CompressToPairs_Default
			( int threshold );
		CompressedDFA_RedundancyRemoved* CompressToRedundancyRemoved();
		int ColumnsAreEquivalent
			(int baseColumn, int inspectedColumn, int originalRowsCount);
		int RowsAreEquivalent
			(int baseRow, int inspectedRow, int originalColumnsCount);
		CompressedDFA_Aho* CompressToAho();
		CompressedDFA* MostCompactTechnique( int threshold );
	};
}




