#include "stdafx.h"
#include <limits>
#include "LexerInputReader.h"
#include <cctype>
#include <algorithm>
#include "Globals.h"
#include "RegularExpressionParser.h"
#include "StreamReader.h"
using namespace std;
using namespace Automata;
//-----------------------------------------------------------------
////// Static Data Definition
const TCHAR *LexerInputReader::ErrorString[ ERROR_TYPE_COUNT ];
const TCHAR *LexerInputReader::ErrorFormatString = _T(" \t%d\tError:%s\t%s\t%d\t ");
//////
void LexerInputReader::InitializeStaticData()
{
	ErrorString[ BAD_DIRECTIVE ] = _T("Bad directive");
	ErrorString[ PREMATURE_EOF ] = _T("Unexpected end of file");
	ErrorString[ INVALID_OPTION ]= _T("Invalid option specification.\
		Proper option format: %option some_thing = some_value");
	ErrorString[ UNKNOWN_OPTION ] = _T("The specified option is not supported");
	ErrorString[ INVALID_MACRO_DEFINITION_FORMAT ] 
			= _T("Invalid macro definition. Eliminate the trailing characters");//name value otherchars_error
	ErrorString[ MACRO_REDIFINITION ] = _T("The specified macro name is already defined");
	ErrorString[ INVALID_REGEX_SPACES ] = _T("Illegal spaces within the regular expression");
	ErrorString[ INVALID_REGEX_PAREN_NOT_BALANCED ] = _T("Invalid use of parentheses within the regular expression. Check balance");
	ErrorString[ INVALID_MACRO_INVOCATION_WITHIN_REGEX ]
							= _T("Invalid macro invocation within the regular expression. The macro name contains invalid characters");
	ErrorString[ UNEXPECTED_END_OF_REGEX_CHECK_BALANCE ] 
							= _T("Unexpected end of regular expression. Check the balance of parentheses");
	ErrorString[ UNEXPECTED_END_OF_REGEX ] = _T("Unexpected end of regular expression");
	ErrorString[ INVOKED_MACRO_IS_UNDEFINED ] = _T("The invoked macro is undefined");
	ErrorString[ REGEX_TERMINATED_WITH_BACKSLASH ] 
							= _T("Unexpected end of regular expression. Remove the trailing backslash");
	ErrorString[ NEGATIVE_RANGE_WITHIN_CCL ] = _T("Negative ranges within a character class are not allowed");
	ErrorString[ UNEXPECTED_END_OF_CCL ] = _T("Unexpected end of character class.");
	ErrorString[ CHAR_OUTSIDE_DEFINED_CHARSET ] = 
		_T("One or more characters within the regular expression are outside the range of the defined character set");
	ErrorString[ BAD_CHARSET_DEFINITION ] = _T("Bad character set definition. Check the supplied %option CharacterSet.");
	ErrorString[ UNKNOWN_ERROR ] = _T("Unknown error");
		
	
}
//////
LexerInputReader::~LexerInputReader(void)
{
	NfaVector::size_type nfaCount = Nfa.size();
	NfaVector::size_type i = 0;
	for(; i < nfaCount; i++)
	{
		if( Nfa[i] != NULL ) delete Nfa[i];
		if( Code[i] != NULL) delete[] Code[i];
	}
}
//////
void LexerInputReader::AddError(ErrorCode ecode, int lineNo)
{
	TCHAR buffer[1024];

	_stprintf(	buffer,
				ErrorFormatString,
				errCount++,
				ErrorString[ ecode ],
				FileName.c_str(),
				lineNo);

	tstring error( buffer );

    Level1ErrorsTable.push_back( buffer );			
}
//////
void LexerInputReader::AddError( ErrorCode ecode )
{
	AddError( ecode, _lineNo );
}
//////
LexerInputReader::LexerInputReader(	InputStream &inputFile,
									Dictionary	&optionTable ) :
	_fin( inputFile ),
	OptionTable( optionTable )
{
	_eof = FALSE;
	_lineNo = 0;
}
//////
LexerInputReader::LexerInputReader(	InputStream &inputFile,
					Dictionary &optionTable,
					const tstring &fileName ) : _fin( inputFile ),
										OptionTable( optionTable ),
										FileName( fileName )
{
	_eof = FALSE;
	_lineNo = 0;
}
//////
void LexerInputReader::CheckEOF()
{
	if( _fin.peek() == char_traits<TCHAR>::eof() || _fin.eof() )
		_eof = TRUE;
	else 
		_eof = FALSE;
}
//////
void LexerInputReader::ReadHeaderTopSection()
{
	tstring		strLine;	//Line		
	BOOL		completed = FALSE;
	TCHAR		c;			//First character of a line.
	int			strLen;		//Length of Line
	long	linePos;//Position of start of line in file.
	strLine.reserve( 1024 );
	CheckEOF();
	while (!completed && ! _eof)
	{
		linePos = _fin.tellg();
		_fin.ReadLine(strLine);
		_lineNo++;
		strLen = strLine.length();
		if(strLen == 0) //Empty line
		{
			HeaderTopSection.append( ENDLINE );
			CheckEOF();
			continue;
		}
		c = strLine[0];
		if( _istspace( c ) )	//white space
		{
			//Eat white spaces and append the rest to 
			//HeaderTopSection
			for(int i = 1; i < strLen; i++)
				if(! _istspace( strLine[i] ) )
				{
					HeaderTopSection.append( strLine.begin() + i,
											 strLine.end() );
					//HeaderTopSection.append(1, _T('\r'));
					HeaderTopSection.append(ENDLINE );
					break;
				}
		}
		else if( c == _T('%') ) //Directive
		{
			if( strLen == 1 )
				AddError( BAD_DIRECTIVE, _lineNo );
			else
			{
				c = strLine[1];
				switch( c )
				{
				case _T('%'): //%%
					completed = true;
					break;
				case _T('{'): //%{
					ReadInsidePercentBraces( strLine );
					break;
				case _T('o'): //%option(w)*pref(w)*=(w)*restOfLine
							  //01234567
					ProcessOptionDirective( strLine );
					break;
				default:
					AddError( BAD_DIRECTIVE, _lineNo );
				}
			}
		}
		else if( c == _T('/') ) //comments
		{
			if( strLen == 1 )
				AddError( UNKNOWN_ERROR, _lineNo );
			else
			{
				c = strLine[1];
				if( c != _T('*') && c != _T('/') )
				{
					AddError( UNKNOWN_ERROR, _lineNo );
				}
				else
				{
					//Return back the line to the input stream
					//_fin.seekg( (- strLen - _tcslen(ENDLINE))*sizeof(TCHAR), ios_base::cur );
					_fin.seekg( linePos );
			//		while( _fin.peek() != _T('/') ) 
			//			_fin.seekg( - (int)sizeof(TCHAR), ios_base::cur );
					CommentType ct = ExtractComment(HeaderTopSection);
					if(	ct == MatchedStarComment ||
						ct == EOFStarComment )
					{
						//Ignore the rest of line.
						while( _fin.get() != _T('\n') ) ;
					}
					HeaderTopSection.append( ENDLINE );
				}
			}
		}
		else if( c == _T('_') || _istalpha(c) ) //Macro
		{
			ProcessMacro(strLine);
		}
		else //Error
		{
			AddError( UNKNOWN_ERROR, _lineNo );
		}
		CheckEOF();
	} //END while (!completed && ! _eof)
	//if( _eof) unexpected end of file ... this left to ParseFile()
}
//////
void LexerInputReader::ReadInsidePercentBraces(tstring &buffer)
{
	CheckEOF();
	while(! _eof )
	{
		_fin.ReadLine( buffer );
		_lineNo++;
		if( buffer.size() >= 2 )
		{
			if( buffer[0] == _T('%') && buffer[1] == _T('}') )
				return; //The sub section %{...%} has been terminated
		}
		//Copy the line to HeaderTopSection
		HeaderTopSection.append( buffer.begin(), buffer.end() );
		HeaderTopSection.append( ENDLINE );
		CheckEOF();
	}
}
//////
void LexerInputReader::ProcessOptionDirective(tstring &line) //This method is implemented inefficiently.
{
	//%option(w)*pref(w)*=(w)*restOfLine
	//01234567
	BOOL isOption = TRUE;
	tstring::size_type offset = 0, offset2;
	tstring::size_type strLen = line.size();

	if( strLen < 10 )
	{
		AddError( BAD_DIRECTIVE, _lineNo );
		return;
	}
	offset = line.find( _T("%option"), offset );
	if(offset != 0) //doesnot begin with %option
	{
		AddError( BAD_DIRECTIVE, _lineNo );
		return;
	}
	offset = 7;
	while( _istspace(line[offset]) ) 
	{
		offset++;
		if( offset >= strLen ) //line contains only %option spaces
		{
			AddError( INVALID_OPTION, _lineNo );
			return;
		}
	}
	if( offset == 7 ) //No separating spaces
	{
		AddError( INVALID_OPTION, _lineNo );
		return;
	}
	//Now, offset points to the first non space character after %option
	offset2 = line.find( _T('='), offset ); //Find '=' position
	if( offset2 == tstring::npos ) //No '='
	{
		AddError( INVALID_OPTION, _lineNo );
		return;	
	}

	tstring optionHeader( line.begin() + offset,
						  line.begin() + offset2 );
	TrimEndFromSpaces(optionHeader);	//Remove trailing spaces
	//Capitalize optionHeader
	transform(	optionHeader.begin(),
				optionHeader.end(),
				optionHeader.begin(),
				toupper);
	Dictionary::iterator iter = OptionTable.find( optionHeader );
	if(iter == OptionTable.end()) //Unknown option
	{
		AddError( UNKNOWN_OPTION, _lineNo );
		return;	
	}
	tstring optionValue( line.begin() + offset2 + 1, line.end() );
	TrimStartFromSpaces( optionValue );
	TrimEndFromSpaces( optionValue );
	iter->second = optionValue;
	//The question now is: Should we validate the option value
	//here or leave it to a later phase?
}
//////
void LexerInputReader::TrimStartFromSpaces(tstring &str)
{
	tstring::size_type k = 0, count = str.length();
	while( k < count )
		if( ! _istspace( str[k] ) ) //Not space
		{
			str.erase(0, k); //remove from pos=0, k chars
			return;
		}
		else k++; //otherwise, it is a space
	//The string is all spaces
	str.clear();
}
//////
void LexerInputReader::TrimEndFromSpaces(tstring &str)
{
	tstring::size_type k = str.size() - 1;
	while( k >= 0 )
		if( ! _istspace( str[k] ) ) //Not space
		{
			str.erase(k+1); //remove from k+1 to end
			return;
		}
		else k--; //otherwise, it is a space
	//The string is all spaces
	str.clear();
}
//////
enum CommentStateType
{
	NotComment,
	BeginSlash,
	LineComment,
	AllComment,
	EndStar,
	EndComment
};
LexerInputReader::CommentType LexerInputReader::ExtractComment(tstring &text)
{
	TCHAR				c = (TCHAR)_fin.get();
	CommentType			commType = None;
	CommentStateType	state = NotComment;
	while(c != char_traits<TCHAR>::eof())
	{
		switch(state)
		{
		case NotComment:
			if(c == _T('/'))
				state = BeginSlash;
			text += c;
			break;
		case BeginSlash:
			switch(c)
			{
			case _T('/'):
				state = LineComment;
				break;
			case _T('*'):
				state = AllComment;
				break;
			default: // It is guaranteed that c = * or / (as a precondition). So, this is useless.
				state = NotComment;
				AddError( UNKNOWN_ERROR, _lineNo );
				if( c == _T('\n') ) _lineNo++;
				return None;
			}
			text += c;
			break;
		case LineComment:
			if(c == _T('\n'))
			{
				commType = MatchedLineComment;
				state = EndComment;
			}
			else 
				text += c;
			break;
		case AllComment:
			if(c == _T('*'))
				state = EndStar;
			else if( c == _T('\n') )
				_lineNo++;
			text += c;
			break;
		case EndStar:
			if(c == _T('/'))
			{
				commType = MatchedStarComment;
				state = EndComment;
			}
			else if(c == _T('\n'))
			{
				_lineNo++;
				state = AllComment;
			}
			else if(c != _T('*'))
				state = AllComment;
			text += c;
			break;
		case EndComment:
//			text += _T('\n');
			return commType;
		}
		if(state != EndComment)
			c = (TCHAR)_fin.get();
	}
	if(c == char_traits<TCHAR>::eof())
	{
		if( state == AllComment )	commType = EOFStarComment;
		else if( state == LineComment )	commType = EOFLineComment;
		_eof = TRUE;
	}
	return commType;
}
//////
int LexerInputReader::ExtractIdentifier(tstring &line,
										tstring &identifier) 
{
	int pos = 0;
	TCHAR c = line[0];
	//int strLen = identifier.size();	That is the error
	int strLen = line.size();
	if( !_istalpha(c) && c != _T('_') ) return pos;
	identifier.append(1, c);
	if( ++pos >= strLen ) return pos;
	c = line[pos];
	while( _istalnum(c) || c == _T('_') )
	{
		identifier.append( 1, c );
		if( ++pos >= strLen ) return pos;
		c = line[pos];
	}
	return pos;
}
//////
void LexerInputReader::ProcessMacro(tstring &line)
{
	//MacroName MacroValue
	tstring			macroName;
	tstring			macroValue;
	int				pos1;
	int				strLen = line.length();//same as line.size()
	tstrIterator	iter, end;
	ErrorCode		errCode;

	pos1 = ExtractIdentifier( line, macroName );
	//if( pos1 == 0 ) Cannot happen since line[0] = [_a-zA-Z] as a precondition.
	if ( _macroTable.find( macroName ) != _macroTable.end() ) //Macro already exists...
	{
		AddError( MACRO_REDIFINITION, _lineNo );
		return;
	}
	if( pos1 == strLen ) // The line consists of only a macro name.
	{
		//Warn the user that the macro will be defined as a single space.
		//Warning
		//Add macro to macroTable
		macroValue.assign( 1, _T(' ') );
		_macroTable[ macroName ] = macroValue;
		return;
	}
	if( ! _istspace( line[pos1] ) ) //If the first char after the 
	{							//macro name is not a whitespace
		AddError( UNKNOWN_ERROR, _lineNo );
		return;
	}
	//skip white spaces
	end = line.end();
	iter = find_if( (tstrIterator)line.begin() + pos1, 
					end, 
					IsNotTSpace); //returns the first nonspace character
	if( iter == end ) //The line is all spaces after the macro 
	{					//name
		//Warn the user that the macro will be defined as a single space.
		//Warning
		//Add macro to macroTable
		macroValue.append( 1, _T(' ') );
		_macroTable[ macroName ] = macroValue;
		return;
	}
	//Extract the regular expression
	iter = ExtractRegularExpression(iter, end, errCode, macroValue);
	if( errCode != NO_ERROR ) //If we faced an error
	{
		AddError( errCode, _lineNo );
		return;
	}
    //Check wether the rest of the string are white spaces or not
	iter = find_if( iter,
					end,
					IsNotTSpace); //returns the first nonspace character
	if( iter < end ) //There is a nonwhite space character before end
	{
		AddError( INVALID_MACRO_DEFINITION_FORMAT, _lineNo );
		return;
	}
	//Add macro (name, value) to the _macroTable
	_macroTable[ macroName ] = macroValue;
}
//////
enum RegexExtractionState
{
	stStart,	//represents normal state as well.
	stInQuotes,
	stInCCL,
	stBadRule,
	stMacroStart,
	stInMacro,
	stFinal
};
tstrIterator LexerInputReader::ExtractRegularExpression(
								tstrIterator	iter,
								tstrIterator	end,
								ErrorCode		&errCode,
								tstring			&regex)
{
	//*iter is not a space
	//iter < end
	//[iter, end) doesnot contain endline (We shall ignore that)
	TCHAR c;
	RegexExtractionState	state = stStart;
	tstrIterator			begin = iter;
	tstrIterator			macroBegin, macroEnd; //MacroName = [macroBegin, macroEnd)
	int						parenBalance = 0; //++ if (, -- if )
	//parenBalance just always remain >= 0.

	Dictionary::const_iterator	macroEntry; //points to the 
						//location of the macro in the macroTable.
	errCode = NO_ERROR;
	while( iter < end ) //[iter, end)
	{
		c = *iter++;
		switch( state )
		{
		case stStart:
			switch(c)
			{
			case _T('\"'): // "
				state = stInQuotes;
				break;
			case _T('\\'): // Escape Sequence
				if( iter == end ) // backslash is last char
				{
					errCode = REGEX_TERMINATED_WITH_BACKSLASH;
					return iter;
				}
				c = *iter++;
				//if( c == _T('\n') ) impossible
				//No checks on c.
				break;
			case _T('['):	//CCL_BEGIN
				state = stInCCL;
				break;
			case _T('('):	// (r)
				parenBalance++;
				break;
			case _T(')'):
				if( parenBalance <= 0 ) //Invalid closing paren
				{
					errCode = INVALID_REGEX_PAREN_NOT_BALANCED;
					return iter;
				}
				parenBalance--;
				break;
			case _T('{'):
				state = stMacroStart;
				break;
			default:
				if( _istspace(c) ) // c is a space
				{
					//The regex should terminate just before 
					//the white space. iter points to the
					//character after the white space.
					//regex c
					//b     i
					iter -= 2;
					//regex c
					//b   i 
					//[    )
					state = stFinal;
					//Note: iter will be incremented before entering
					//stFinal. So, the behaviour will be the desired
					//one.
				}
				//else do nothing
			}
			break;
		case stMacroStart:
			if( c != _T('_') && ! _istalpha(c) ) //not [_a-zA-Z]
			{
				errCode = INVALID_MACRO_INVOCATION_WITHIN_REGEX;
				return iter;
			}
			//[_a-zA-Z]
			macroBegin = iter-1;//macroBegin points to the first
								//char in macro name
			state = stInMacro;
			break;
		case stInMacro:
			if( c == _T('}') ) // Macro ended
			{
				//iter points to the char just after }
				macroEnd = iter - 1; //macroEnd points to }
				//Check wether the invoked macro is defined
				macroEntry = _macroTable.find( 
						tstring( macroBegin, macroEnd ) );
				if( macroEntry == _macroTable.end() ) //Undefined macro
				{
					errCode = INVOKED_MACRO_IS_UNDEFINED;
					return iter;
				}
				//otherwise, continue working!
				state = stStart;
				/* 
				Since we check at macro definition, wether the 
				invoked macro is already defined, we guarantee
				macro references are all forward. So, no recursion
				or circulation occurs.
				*/
			}
			else if( c != _T('_') && ! _istalnum(c) ) // not [_0-9a-zA-Z]
			{
				//invalid macro name
				errCode = INVALID_MACRO_INVOCATION_WITHIN_REGEX;
				return iter;
			}
			//else we are still in macro
			break;
		case stInCCL:
			if( c == _T(']') ) //CCL_END
				state = stStart;
			else if( c == _T('\\') ) //Esacape Sequence
			{
				if( iter == end ) // backslash is last char
				{
					errCode = REGEX_TERMINATED_WITH_BACKSLASH;
					return iter;
				}
				c = *iter++;
				//if( c == _T('\n') ) impossible
				//No checks on c.
			}
			//else 	//if( c == _T('\n') ) impossible
			//else stay on the same state
			break;
		case stInQuotes:
			if( c == _T('\"') ) // END_QUOTES
				state = stStart;
			else if( c == _T('\\') ) //Esacape Sequence
			{
				if( iter == end ) // backslash is last char
				{
					errCode = REGEX_TERMINATED_WITH_BACKSLASH;
					return iter;
				}
				c = *iter++;
				//if( c == _T('\n') ) impossible
				//No checks on c.
			}
			//else if( c == _T('\n') ) impossible
			//else stay in the same state.
			break;
		case stFinal:
			if( parenBalance != 0 )
			{
				AddError( INVALID_REGEX_SPACES, _lineNo);
				return iter;
			}
			regex.assign( begin, iter );
			return iter;
		} // End switch( state )
	}//End while( iter < end )
	//We come here only if iter = end. This means we reached end of string
	if( state == stStart ) //stStart --end of string--> stFinal
	{
		if( parenBalance == 0 )
		{	
			regex.assign( begin, iter );
			return iter;
		}
		else
		{
			errCode = INVALID_REGEX_PAREN_NOT_BALANCED;
			return iter;
		}
	}
	else //Unmatched " or [ or {. This depends wether we
		//state = stInQuotes or stInCCL or stInMacro or stMacroStart
	{
		errCode = UNEXPECTED_END_OF_REGEX;
		return iter;
	}

}
//////
void LexerInputReader::ReadClassDefinition()
{
	tstring		strLine;	//Line
	BOOL		completed = FALSE;
	strLine.reserve( 1024 );
	while( !completed && !_eof )
	{
		_fin.ReadLine( strLine );
		_lineNo++;
		if( strLine.length() < 1 )
			ClassDefinition.append( ENDLINE );
		else if( strLine.find( _T("%%"), 0 ) == 0 ) //starts with %%
		{
			completed = TRUE;
			//we may also check if the line contains trailing 
			//invalid characters.
		}
		else //Append the line
		{
			ClassDefinition.append( strLine );
			ClassDefinition.append( ENDLINE );
		}
		CheckEOF();
	}
	//if _eof : premature end of file. We leave to ParseFile()
}
//////
void LexerInputReader::ReadCppTopSection()
{
	tstring		strLine;	//Line
	BOOL		completed = FALSE;
	strLine.reserve( 1024 );
	while( !completed && !_eof )
	{
		_fin.ReadLine( strLine );
		_lineNo++;
		if( strLine.length() < 1 )
			CppTopSection.append( ENDLINE );
		else if( strLine.find( _T("%%"), 0 ) == 0 ) //starts with %%
		{
			completed = TRUE;
			//we may also check if the line contains trailing 
			//invalid characters.
		}
		else //Append the line
		{
			CppTopSection.append( strLine );
			CppTopSection.append( ENDLINE );
		}
		CheckEOF();
	}
	//if _eof : do nothing. The file might end here.
}
//////
void LexerInputReader::ReadCppBottomSection()
{
	tstring		strLine;//Line
	BOOL		completed = FALSE;
	strLine.reserve( 1024 );
	while( !completed && !_eof )
	{
		_fin.ReadLine(strLine);
		_lineNo++;
		if( strLine.length() < 1 )
			CppBottomSection.append( ENDLINE );
		else if( strLine.find( _T("%%"), 0 ) == 0 ) //starts with %%
		{
			completed = TRUE;
			//we may also check if the line contains trailing 
			//invalid characters.
		}
		else //Append the line
		{
			CppBottomSection.append( strLine );
			CppBottomSection.append( ENDLINE );
		}
		CheckEOF();
	}
	//if _eof : do nothing. The file might end here.
}
//////
void LexerInputReader::ReadMachineDescription()
{
	tstring			strLine;//Line
	int				strLen;	//Length of Line
	tstrIterator	iter, end; //Line iterators
	BOOL			completed = FALSE;
	TCHAR			c; //First character of a line.
	NFA				*pNfa;		//Pointer to NFA
	tstring			action;//Code Action
	tstring			regex;
	ErrorCode		errCode;

	strLine.reserve( 1024 );
	action.reserve( 1024 );
	regex.reserve( 128 );

	RegularExpressionParser regexParser( this );
	CheckEOF();
	while (!completed && ! _eof)
	{
		_fin.ReadLine( strLine );
		_lineNo++;
		strLen = strLine.length();
		if(strLen == 0) //Empty line
		{
			CheckEOF();
			continue;
		}
		c = strLine[0];
		if( _istspace( c ) )	//white space: Indented line
		{
			//Ignore the indented line
			CheckEOF();
			continue;
		}
		else if( c == _T('%') && strLen > 1 &&
				 strLine[1] == _T('%') )//%% Terminator
			completed = true;
		else //Rule: Regex Action
		{
			iter = strLine.begin();
			end = strLine.end();
			//extract the regular expression
			iter = ExtractRegularExpression(iter,
											end,
											errCode,
											regex);
			if( errCode != NO_ERROR ) //There was an error
			{
				AddError(errCode, _lineNo);
			}
			else
			{
				pNfa = new NFA( &_mem );
				action.clear();
				//Since the parser adds the appropriate errors to the errors
				//table in case of errors, we donot need to check the return
				//value. However, I do check it so that I donot add it the
				//NfaVector in case of errors. (I can just check the errors
				//table for any errors and add only if no errors.)
				regexParser.Parse( regex, *pNfa );
				ExtractActionCode( iter, end, action );
				if( Level1ErrorsTable.size() == 0 )
				{
					Nfa.push_back( pNfa );
					Code.push_back( CopyString( action.c_str() ) );
				}
				else
				{
					delete pNfa;
				}
			}
		}
		CheckEOF();
	} //END while (!completed && ! _eof)

}
///////
BOOL LexerInputReader::ParseFile()
{
	//Range of the character set.
	TCHAR			startSymbol, finalSymbol;
	ErrorCode		errCode;

	ReadHeaderTopSection();
	if( _eof ) //If file has only the header top section
	{
		AddError( PREMATURE_EOF, _lineNo );
		return FALSE;
	}
	//If there are errors, should we terminate or proceed?
	//if( Level1ErrorsTable.size() > 0 ) return FALSE;

	ReadClassDefinition();

	if( _eof ) //If file has no rule section
	{
		AddError( PREMATURE_EOF, _lineNo );
		return FALSE;
	}

	//Now, this is the time to determine the range of the
	//character set.
	errCode = DecideCharacterRange( startSymbol, finalSymbol );
	if(errCode != NO_ERROR)
	{
		Level1ErrorsTable.push_back( 
			tstring( ErrorString[ BAD_CHARSET_DEFINITION ] )
			+ tstring( _T(" Details:") )
			+ tstring( ENDLINE )
			+ tstring(ErrorString[ errCode ])
			+ tstring(_T("."))	);
		return FALSE;// We cannot parse the regular expressions
				//with invalid character set.
	}
	_faFactory.Reset( startSymbol, finalSymbol );
	//Construct the memory manager
	_mem.ReInitialize( &_faFactory );

	//Read the rules section.
	ReadMachineDescription();
	
	if( _eof ) //If end of file, the remaining sections are optionals
		return Level1ErrorsTable.size() == 0;
	//Since there is no further error checking, read the remaining 
	//sections only if no errors have occured
	if( Level1ErrorsTable.size() > 0 ) return FALSE;
	//Read the CPP Top Section
	ReadCppTopSection();
	
	if( _eof )//If end of file, return TRUE
		return TRUE;

	//Read the CPP Bottom Section
	ReadCppBottomSection();

	return TRUE;
}
//////
enum CCLState
{
	cclStart,
	cclFinal,		//cclFinal: Preconds: iter points to the
					// first char after ]. It just returns TRUE
					//It is found that going to cclFinal causes
					//many errors. So, we shall ignore going there.
	cclAcceptance, //Normal in documentation
	cclRejection,
	cclAcceptanceRange,
	cclRejectionRange
};
LexerInputReader::ErrorCode LexerInputReader::ParseCharacterClass(	
								tstrIterator		&iter,
								const tstrIterator &end,
								BOOL			onlyPositiveRanges,
								RangedSet&		charSet)
{
	TCHAR		c, storedChar, lookAhead;
	BOOL		stored; //stored and storedChar are used to handle
						//CCL ranges properly. For further info.
						//about how they work, see the machine
						//in the documentation and note that
						//stored = -1 means stored = false.
						//That is, there is no stored char.
	CCLState	state;
	iter++;		//skip CCL_BEGIN
	state = cclStart;
	//At cclFinal, we return TRUE. So, if the ccl is correct,
	//this loop will never terminate
	while( iter < end )
	{
		c = *iter;	//As a convention here, iter will always point 
					//at the current char c.
		switch(state)
		{
		case cclStart:
			switch(c)
			{
			case _T('^'): //[^
				//Make a lookahead to check for the degenerate
				//case [^]
				lookAhead = *(iter+1);
				if( lookAhead == _T(']') ) //[^]: Accept just ^
				{
					//[^]
					// i      need to advance after ']'
					iter += 2;
					//[^]
					//   i
					if( ! charSet.WithinRange(c) ) 
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Add( c ); //_T('^')
					//state = cclFinal; //End
					return NO_ERROR;
				}
				else //It is a rejection class
				{
					stored	= FALSE; //No initial range
					charSet.AddAllElements();
					state	= cclRejection;
				}
				break;
			case _T('\\'): //Escape sequence
				//since we are sure that at least we have
				//[\], it is safe to call ParseEscapeSequence
				iter++; //iter should point after backslash
				storedChar = ParseEscapeSequence(iter, end);
				//since iter points after the last char in
				//the escape sequence, we need to decrement
				//iter since it will be incremented again.
				iter--;
				stored = TRUE;
				state = cclAcceptance;
				break;
			default: //Any char other than \, ^
				storedChar = c;
				stored = TRUE;
				state = cclAcceptance;
				break;
			}
			break;
		case cclAcceptance:
			switch(c)
			{
			case _T('\\'): //Escape sequence
				if( stored )
				{
					if( ! charSet.WithinRange( storedChar ) )  
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Add( storedChar );
					//stored = FALSE; //USELESS
				}
				iter++; //iter should point after backslash
				//Preconds: iter < end
				if( iter >= end ) // terminated by backslash
					return UNEXPECTED_END_OF_CCL;
				storedChar = ParseEscapeSequence(iter, end);
				//since iter points after the last char in
				//the escape sequence, we need to decrement
				//iter since it will be incremented again.
				iter--;
				stored = TRUE;
				//state = cclAcceptance; //Useless
				break;
			case _T('-'): //if(stored) then range else char
				if( stored ) // Go to range
				{
					state = cclAcceptanceRange;
				}
				else
				{
					storedChar = c; //_T('-')
					stored = TRUE;
					//state = cclAcceptance; //Useless
				}
				break;
			case _T(']'): //CCL_END
				if( stored ) //There is a stored char
				{
					if( ! charSet.WithinRange( storedChar ) )
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Add( storedChar );
					stored = FALSE;
				}
				iter++; //iter must point just after ] before
						//travelling to the final state
				//state = cclFinal; // Instead of going to cclFinal, 
				//return immediately
				return NO_ERROR;
				//break;
			default: //Any char other than ], -, '\'
				if( stored )
				{
					if( ! charSet.WithinRange( storedChar ) ) 
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Add( storedChar );
					//stored = FALSE; //USELESS
				}
				storedChar = c;
				stored = TRUE;
				//state = cclAcceptance; //Useless
				break;
			}
			break;
		case cclAcceptanceRange:
			switch(c)
			{
			case _T('\\'): //Escape sequence
				iter++; //iter should point after backslash
				//Preconds: iter < end
				if( iter >= end ) // terminated by backslash
					return UNEXPECTED_END_OF_CCL;
				c = ParseEscapeSequence(iter, end);
				//since iter points after the last char in
				//the escape sequence, we need to decrement
				//iter since it will be incremented again.
				iter--;
				//Now, add the range [storedChar - c]
				if( ! charSet.WithinRange( storedChar ) || 
					! charSet.WithinRange( c ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				if( onlyPositiveRanges && storedChar > c )
					return NEGATIVE_RANGE_WITHIN_CCL;
				//Check both storedChar and c against maxElement
				charSet.AddRange( storedChar, c );
				stored = FALSE;
				state = cclAcceptance;
				break;
			case _T(']'): //CCL_END.
                //Add storedChar and -
				if( ! charSet.WithinRange( storedChar ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				charSet.Add( storedChar );
				stored = FALSE;
				if( ! charSet.WithinRange( _T('-') ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				charSet.Add( _T('-') );
				iter++; //iter must point just after ] before
						//travelling to the final state
				//state = cclFinal; // Instead of going to cclFinal, 
				//return immediately
				return NO_ERROR;
				//break;
			default: //Any char other than ], '\'
				//Of-course, stored = TRUE
				//Range [storedChar - c]
				if( !charSet.WithinRange( storedChar ) ||
					!charSet.WithinRange( c ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				if( onlyPositiveRanges && storedChar > c )
					return NEGATIVE_RANGE_WITHIN_CCL;
				charSet.AddRange( storedChar, c );
				//The previous two statements could have been
				//merged into: 
				//if(! charSet.AddRange(storedChar, c) )
				//		return CHAR_OUTSIDE_DEFINED_CHARSET
				stored = FALSE;
				state = cclAcceptance;
				break;
			}
			break;
		case cclRejection:
			switch(c)
			{
			case _T('\\'): //Escape sequence
				if( stored )
				{
					if( ! charSet.WithinRange( storedChar ) )  
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Remove( storedChar );
					//Could be combined: if( !charSet.Remove( c ))
					//						return error...
					//stored = FALSE; //USELESS
				}
				iter++; //iter should point after backslash
				//Preconds: iter < end
				if( iter >= end ) // terminated by backslash
					return UNEXPECTED_END_OF_CCL;
				storedChar = ParseEscapeSequence(iter, end);
				//since iter points after the last char in
				//the escape sequence, we need to decrement
				//iter since it will be incremented again.
				iter--;
				stored = TRUE;
				//state = cclRejection; //Useless
				break;
			case _T('-'): //if(stored) then range else char
				if( stored ) // Go to range
				{
					state = cclRejectionRange;  //This was an error
												//it was written
												//cclAcceptanceRange
				}
				else
				{
					storedChar = c; //_T('-')
					stored = TRUE;
					//state = cclRejection; //Useless
				}
				break;
			case _T(']'): //CCL_END
				if( stored ) //There is a stored char
				{
					if( ! charSet.WithinRange( storedChar ) )
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Remove( storedChar );
					stored = FALSE;
				}
				iter++; //iter must point just after ] before
						//travelling to the final state
				//state = cclFinal; // Instead of going to cclFinal, 
				//return immediately
				return NO_ERROR;
				//break;
			default: //Any char other than ], -, '\'
				if( stored )
				{
					if( ! charSet.WithinRange( storedChar ) ) 
						return CHAR_OUTSIDE_DEFINED_CHARSET;
					charSet.Remove( storedChar );
					//stored = FALSE; //USELESS
				}
				storedChar = c;
				stored = TRUE;
				//state = cclRejection; //Useless
				break;
			}
			break;
		case cclRejectionRange:
			switch(c)
			{
			case _T('\\'): //Escape sequence
				iter++; //iter should point after backslash
				//Preconds: iter < end
				if( iter >= end ) // terminated by backslash
					return UNEXPECTED_END_OF_CCL;
				c = ParseEscapeSequence(iter, end);
				//since iter points after the last char in
				//the escape sequence, we need to decrement
				//iter since it will be incremented again.
				iter--;
				//Check both storedChar and c against maxElement
				if( ! charSet.WithinRange( storedChar ) || 
					! charSet.WithinRange( c ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				//Now, remove the range [storedChar - c]
				if( onlyPositiveRanges && storedChar > c )
					return NEGATIVE_RANGE_WITHIN_CCL;
				charSet.RemoveRange( storedChar, c );
				stored = FALSE;
				state = cclRejection;
				break;
			case _T(']'): //CCL_END.
                //Remove storedChar and -
				if( ! charSet.WithinRange( storedChar ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				charSet.Remove( storedChar );
				stored = FALSE;
				if( ! charSet.WithinRange( _T('-') ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				charSet.Remove( _T('-') );
				iter++; //iter must point just after ] before
						//travelling to the final state
				//state = cclFinal; // Instead of going to cclFinal, 
				//return immediately
				return NO_ERROR;
				//break;
			default: //Any char other than ], '\'
				//Of-course, stored = TRUE
				//Range [storedChar - c]
				if( onlyPositiveRanges && storedChar > c )
					return NEGATIVE_RANGE_WITHIN_CCL;
				if( !charSet.WithinRange( storedChar ) ||
					!charSet.WithinRange( c ) )
					return CHAR_OUTSIDE_DEFINED_CHARSET;
				charSet.RemoveRange( storedChar, c );
				//The previous two statements could have been
				//merged into: 
				//if(! charSet.RemoveRange(storedChar, c) )
				//		return CHAR_OUTSIDE_DEFINED_CHARSET
				stored = FALSE;
				state = cclRejection;
				break;
			}
			break;
		case cclFinal:
			return NO_ERROR;
			break;
		}//end switch(state)
		iter++; //Advance
	}//end while( iter < end )
	
	//Since we reached here, iter = end and we didnot see ]
	return UNEXPECTED_END_OF_CCL;
}
//////
LexerInputReader::ErrorCode LexerInputReader::DecideCharacterRange(
										TCHAR &startSymbol,
										TCHAR &finalSymbol)
{
	ErrorCode		errCode;
	tstrIterator	cclIter, cclEnd;
	int				temp;
	BitSet			charSet( std::numeric_limits<TCHAR>::max() -
		std::numeric_limits<TCHAR>::min() );
	RangedSet		rangedCharSet( &charSet,
									numeric_limits<TCHAR>::min());
	//Now, this is the time to determine the range of the
	//character set.
	//Set default values
	startSymbol = START_SYMBOL;
	finalSymbol = FINAL_SYMBOL;
	//Get charSet string from the OptionTable
	//Note: Wether the user has or has not specified the char
	//set, all Option Names are present with default values
	//in the OptionTable. So, the following statement cannot
	//fail.
	const tstring &charSetStr = OptionTable[OPTION_CHARACTER_SET];
	cclIter = charSetStr.begin();
	cclEnd = charSetStr.end();
	if( cclIter == cclEnd ) //empty string, although I think that 
					//the parser doesnot allow empty option values.
		  return NO_ERROR; //return default char set
		
	while( _istspace( *(cclEnd-1) ) ) //skip trailing white spaces
	{
		cclEnd--;
		if( cclEnd <= cclIter ) //If all of the string are white spaces
			return NO_ERROR;  //return default char Set
				//Although I think the parser doesnot allow this.
	}

	while( _istspace ( *cclIter ) ) //skip leading white spaces
							   //( although supposed to be eliminated )
	{
		cclIter++;
		if( cclIter >= cclEnd ) //Empty string, (suppose not parser
						  //doesnot allow this)
		  return NO_ERROR; //return default char set. 
		//iter >= end: Can never occur here.
		//We shall use a proof by contradication:
		//Suppose that iter gets >= end here. This means that
		//!(iter >= end) before coming here  <-->  iter < end
		//before coming here. So, in the previous loop, iter 
		//has stopped passing through white spaces in the direction
		//of end because of a nonwhite space character. This
		//contradicts the assumption that the string is all spaces.
	}
	//Precond: *cclIter = [, cclIter < cclEnd-2
	if( *cclIter != _T('[') )
	{
		return BAD_CHARSET_DEFINITION;
	}
	//if( cclIter >= cclEnd - 2 ) //This causes problems with security in C++ 2005
	if( cclEnd - cclIter <= 2 )
	{
		//Check if *(cclEnd-1) == _T(']'): So that ccl = []
		if( *(cclEnd-1) == _T(']') )
		{
			return NO_ERROR; //Consider [] [\d0-\d127]
		}
		else //Either ccl = [ or ccl = ['invalidChar'
		{
			return UNEXPECTED_END_OF_CCL;
		}
	}
	errCode = ParseCharacterClass(	cclIter, 
									cclEnd, 
									FALSE,
									rangedCharSet );
	if(errCode == NO_ERROR)
	{
		//There is still some error check!
		//Check that cclIter == cclEnd
		if(cclIter != cclEnd)
			return BAD_CHARSET_DEFINITION;
//		startSymbol = charSet.Minimum(); Wrong!!! You should call
//										 that of RangedSet
//		finalSymbol = charSet.Maximum();
		if( rangedCharSet.MinimumElement( temp ) ) //If there is at
												  //least one element
		{
			startSymbol = temp;
			rangedCharSet.MaximumElement( temp );
			finalSymbol = temp;
		}
	}
	return errCode;
}

//////
enum	ActionCodeState
{
	acNormal,	//Didnot start with %{
	acStarComment,
	acStarCommentPossibleEnd, //saw *
//	acFinal,
	acBracketNormal,//if code started with %{
	acBracketPossibleEnd, //saw %, and ready to stop at }
	// %{.....%?
//	acBracketFinal	//After mathing %}
};
void LexerInputReader::ExtractActionCode(
							tstrIterator iter,
							tstrIterator end, 
							tstring		 &actionCode)
{
	/*
	Algorithm:
		
	*/
	TCHAR			c;
	int				parentheses = 0;	//Number of open parenthes
	BOOL			completed = FALSE;
	BOOL			success; //for checking return values
	ActionCodeState state = acNormal; //Assume normality
	tstrIterator	tempIter;
	tstring			strLine;
	strLine.reserve( 1024 );

	if(iter == end) return;

	//Consume the leading whitespaces
	while( _istspace( *iter ) )
	{
		actionCode.append( 1, *iter++ );
		if( iter >= end ) //all spaces
			return;
	}
	//Check wether it starts with %{
	if( *iter == _T('%') )
	{
		iter++; //Advance
		if( iter == end ) //The action is just %
		{
			actionCode.append( 1, _T('%') ); //append
			return;
		}
		if( *iter == _T('{') ) //it starts with %{
		{
			iter++; //point to the letter after %{
			state = acBracketNormal;
		}
		else //not started with %{
		{
			iter--; //repoint to %
		}
	}

	while(!completed)
	{
		while(iter < end) //While ! end of line
		{
			c = *iter;
			switch(state)
			{
			case acNormal:
				switch(c)
				{
				case _T('{'): //begin code block
					parentheses++;
					actionCode += c;
					break;
				case _T('}'): //end code block
					parentheses--;
					actionCode += c;
					break;
				case _T('/'): //We may encounter a comment
					if( iter + 1 >= end ) //if line ends with /
						break; //No comment Is Possible.
					actionCode += c; //append /
					//Make a look ahead
					if( *(iter+1) == _T('/') ) //Line comment
					{
						iter++; //iter points to the second /
						//append the rest of the line.
						actionCode.append( iter, end );
						iter = end - 1; //So that it is incremented
										//to reach end
					}
					else if( *(iter+1) == _T('*') ) //Star Comment
					{
						iter++; //let iter point to *
						actionCode.append( 1, *iter );
						state = acStarComment;	//Inside comment
					}
					//else no comments
					break;
				case _T('\"'): //"..."
					//Note: it is easy to augment the current
					//machine with the " " machine and even
					//safer.
					actionCode.append( 1, c );
					if(iter + 1 >= end) //ends in "
						break; //No quoted string
					//Precond: iter points after "
					iter++;
					success = MatchQuotedText( 
									iter,
									 end,
									FALSE, //do not convert escs
									TRUE, //permit endlines
									actionCode); //append to action Code
                   // if(! success) //no success if we reached
									//end without closing "
					//However, we donot care even if it was
					//ended with backslash. Many languages such
					//as Java and C# doesnot support extending
					//comments through lines
					if(success) //append closing "
					{
						actionCode += *iter;
					}
					break;
				default: //Not comment, nor { }, nor " "
					actionCode += c;
					break;
				}
				break;
			case acStarComment:	// /*.....?
				if( c == _T('*') )
					state = acStarCommentPossibleEnd;
				actionCode += c;
				break;
			case acStarCommentPossibleEnd: // /*......*?
				if( c == _T('/') )
					state = acNormal;
				//if c == _T('*'), remain on the same state
				else if( c != _T('*') )
					state = acStarComment;
				actionCode += c;
				break;
			case acBracketNormal: //inside %{....?
				if( c == _T('%') )
					state = acBracketPossibleEnd;
				actionCode += c;
				break;
			case acBracketPossibleEnd: //%{......%?
				if( c == _T('}') ) //%{.....%} end now
				{
					//remove % from the end of actionCode
					actionCode.erase( --actionCode.end() );
					//state = acBracketFinal;
					return;
				}
				//else if( c == _T('%') ) , remain at the same state
				else if(c != _T('%') )
				{
					state = acBracketNormal;
				}
				actionCode += c;
				break;
//				case acBracket
			}//END switch(state)
			iter++;
		}//END while(! completed)
		//append endline
		actionCode.append( ENDLINE );
		//Read a new line only if:
		//a)state = acNormal and parentheses > 0
		//b)state = acStarComment or acStarCommentPossibleEnd
		//c)state = acBracket*
		//ie, donot read a new line only if state = acNormal
		//and parentheses <= 0
		if(state == acNormal && parentheses <= 0)
			completed = TRUE;
		else //read a new line (if possible)
		{
			CheckEOF();
			if( _eof ) //Although there might be incomplete matched
					//code (thus possible errors), we donot care.
			{
				completed = TRUE;
			}
			else //have not reached end of file
			{	
				_fin.ReadLine( strLine );
				_lineNo++;
				iter = strLine.begin();
				end = strLine.end();
			}
		}
	}
}