#pragma once
#include "stdafx.h"
#include "NFA.h"
#include "RangedSet.h"
using namespace std;
using namespace Automata;
//The object should be kept alive until the the DFA has been
//constructed out of the NFA and Code vectors. If it is dead 
//before, the NFAs' states will point to NULLs.
namespace Automata
{
class LexerInputReader
{
	friend class RegularExpressionTokenizer;
	friend class RegularExpressionParser;
	/*
	Notes:
	a) Unlike lex, we do not copy indented lines in the machine
	description section into the generated file.
	
	*/
	/*
	A possible problem is the end of line. In MS Windows, end of
	line is the sequence of characters "\r\n" not just "\n".
	So, we may need to define a function that appends end of line
	to a tstring.

	Another problem is with EOF. We ought to use 
	char_traits<TCHAR>::eof() rather than EOF. However, this is
	not really a problem since -1 is the same when compared with
	between int and byte or short and byte. The sign bit is
	extended to the left.
	*/
public:
	Dictionary		_macroTable;
	BOOL			_eof;
	InputStream&	_fin;
	FAFactory		_faFactory;
	NFA::MemoryManager _mem;	//NFA memoryManager. Initially, the character set will not be assigned. After the first section, however, we can construct it.
								//We have choosen to initialize it at the beginning of the Machine section.
	int				_lineNo;	//The number of lines read so far.
public:
	tstring			FileName;	//The input file name. Useful for 
								//error handling
/***Automata***/
	NfaVector		Nfa;	//Nfa vector constructed from the 
							//input file by Thompson Construction.
							//The Nfas are deallocated at the 
							//destructor.
	
	CodeVector		Code;	//Code vector read from the input file.
							//The codes are deallocated at the 
							//destructor.
	Dictionary&		OptionTable;//Option Table supplied as an 
								//input to the object and altered
								//in the input file.
/***Text Data***/
	tstring			HeaderTopSection;	//Text to be placed at the
										//the top of the header 
										//file.

    tstring			ClassDefinition;	//Text to be placed inside
										//the class.
	
	tstring			CppTopSection;	
								//C++:	Text to be placed at the
								//		top of .cpp
								//C#/Java:	Text to be placed
								//			outside the class but
								//			inside the namespace.
	
	tstring			CppBottomSection;
							//C++:		Text to be placed at the
							//			bottom of .cpp
							//C#/Java:	Text to be placed
							//			outside the namespace.
	
public:
	enum ErrorCode;
	enum CommentType
	{
		MatchedLineComment,//      \n
		EOFLineComment, //		EOF
		MatchedStarComment,/*     */
		EOFStarComment,// /*       EOF
		None
	};
/***Methods***/
	LexerInputReader(	InputStream &inputFile,
						Dictionary &optionTable);
	LexerInputReader(	InputStream &inputFile,
						Dictionary &optionTable,
						const tstring &fileName );

	//Reads the input file, extracts rules, and produces the
	//automata, actions, and various code sections. It returns TRUE only
	//if there were no errors.
	BOOL ParseFile();

	/*	Parses the character class located in [iter, end) into 
	charSet and returns NO_ERROR or error code value according to 
	wether the parsing succeeded or failed. In addition, if any 
	character is outside the acceptable range of charSet, the 
	function considers it an error. It stops at the first 
	non-escaped _T(']') (CCL_END). If onlyPositiveRanges = TRUE, 
	then ranges like [z-a] is considered an error and the 
	function fails.	Otherwise, it is treated as if it was [a-z]. 
	Preconditions: 
	a) *iter = _T('[').
	b) iter < end - 2. Ie, length of string is at least 3. 
	   That's, [] is not allowed. (Catastrophic failure occurs 
	   if it was passed such a string). 
	c) charSet is empty. 
	Postconditions: 
	a) If the function returns NO_ERROR, charSet contains the 
	   characters corresponding to the character class [iter, end).
	b) If the function returns NO_ERROR, iter will point the 
	   first character after CCL_END, ].*/
	static ErrorCode	ParseCharacterClass(
								tstrIterator		&iter,
								const tstrIterator	&end,
								BOOL			onlyPositiveRanges,
								RangedSet&		charSet);
public:
	void	CheckEOF();	//Checks wether the next character is EOF. It updates _eof.
	void	ReadHeaderTopSection(); //Reads the HeaderTopSection of the input file.
	void	ReadInsidePercentBraces(tstring &buffer); //Reads and appends section %{...%}. Note: %{ is considered to be read before invocation. buffer is passed so that the function doesnot allocate another buffer.
	void	ProcessOptionDirective(tstring &line);	//Takes a line that supposedly contain %option pref = .*. It process the option and handles the errors.
	void	ProcessMacro(tstring &line); //Extracts and handles a macro. Preconditions: line[0] = [_a-zA-Z]
	//Extracts an identifier [_a-zA-Z][_a-zA-Z0-9]+ from line into 
	//identifier. It returns the index of the next character.
	//Preconditions: 
	//a) line.size() >= 1.
	//b) line[0] = [_a-zA-Z].
	int		ExtractIdentifier(tstring &line, tstring &identifier);

	//It takes the iterators of the current line just after the regular 
	//expression and extracts the actionCode from that line and the
	//following lines (if necessary). It appends the extracted text
	//to actionCode.
	//Preconditions:
	//a) iter <= end
	void	ExtractActionCode(	tstrIterator iter,
								tstrIterator end, 
								tstring		 &actionCode);
	void	AddError( ErrorCode ecode, int lineNo );
	void	AddError( ErrorCode ecode ); //Adds the specified error at the current line number.
	//TrimEnd/StartFromSpaces() should be placed at Templates and should operate on general basic_strings.
	void	TrimEndFromSpaces(tstring &str); //Remove the spaces at the end of the string.
	void	TrimStartFromSpaces(tstring &str);//Remove the spaces at the start of the string.
	CommentType	ExtractComment(tstring &text); //Extracts Cpp Comments or C comments from the InputStream and returns the comment type. _fin.peekString() must be // or /* as a precondition for proper operation and proper _lineNo synchronization. As a postcondition, _fin.peek() is the character just after the end of the comment. This is the beginning of a new line in case of //Comment or the next char to */Comment. It appends the comment to the string text. It returns None if no comment, MatchedStartComment if /* */, MatchedLineComment // \n, EOFStarComment /* EOF, EOFLineComment // EOF
	void	ReadClassDefinition();//Reads the ClassDefinition section terminated with %% from the input file.
	void	ReadCppTopSection();//Reads the CppTopSection terminated with %% or EOF from the input file.
	void	ReadCppBottomSection();//Reads the CppBottomSection terminated with %% or _eof from the input file.
	void	ReadMachineDescription();//Reads the rules section from the input file. It is either ter
	ErrorCode	DecideCharacterRange( TCHAR &startSymbol, TCHAR &finalSymbol );//It determines the range of the character class defined in the character set option in the option table. It returns an ErrorCode corresponding to wether or not the character class was parsed properly.
	

	tstrIterator ExtractRegularExpression(	//Extracts a regular expression from [iter, end) and stores it into regex. It returns the iterator value just after the regex. It sets errCode NO_ERROR if no errors occured. Otherwise, errCode is assigned the appropriate error code. Precondition: a) *iter is not a white-space character. b) [iter, end) are within a single line, ie, there is no endline character in [iter, end). c) iter < end. Postconditions: a) regex contains the regular expression if no error occured. Otherwise, it is left as it is. b) errCode is assigned the appropriate error code. c) return value points just after regex or just after the error.
								tstrIterator iter,
								tstrIterator end,
								ErrorCode	 &errCode,
								tstring &regex);
	~LexerInputReader(void);

public:
/***Error Handling***/
/*
	Until now, error handling is static. We donot provide the user
	enough information. To provide dynamic error handling, it
	is required to add a message format for each error so that
	dynamic error information is provided for each type of 
	error.
	For example, it would be better if we tell the user the 
	sequence of characters to eliminate if we faced the error
	INVALID_MACRO_DEFINITION_FORMAT. It is known that they
	are the characters that follows the macro value till the 
	end of the line. Another example is to tell the macro name
	in case of INVOKED_MACRO_IS_UNDEFINED or 
	INVALID_MACRO_INVOCATION_WITHIN_REGEX.

	We do not handle or even inform the user if he added more
	characters after the section separator. For example,
	%%hello boys hello boys how are you
	is considered a very correct section separator.

	In addition, we need to handle warnings. So, we need to
	add a warning table.
*/
	enum ErrorCode
	{
		NO_ERROR = -1,
		BAD_DIRECTIVE = 0,
		PREMATURE_EOF,
		INVALID_OPTION,
		UNKNOWN_OPTION,
		INVALID_MACRO_DEFINITION_FORMAT, //name value otherchars_error
		MACRO_REDIFINITION,
		INVALID_REGEX_SPACES,
		INVALID_REGEX_PAREN_NOT_BALANCED,
		INVALID_MACRO_INVOCATION_WITHIN_REGEX,
		UNEXPECTED_END_OF_REGEX_CHECK_BALANCE,
		UNEXPECTED_END_OF_REGEX,
		INVOKED_MACRO_IS_UNDEFINED,
		REGEX_TERMINATED_WITH_BACKSLASH,
		UNKNOWN_ERROR,
		NEGATIVE_RANGE_WITHIN_CCL,
		UNEXPECTED_END_OF_CCL,
		CHAR_OUTSIDE_DEFINED_CHARSET,
		BAD_CHARSET_DEFINITION, //This error is added manually in the
								//ParseFile() function without calling
								//AddError
		ERROR_TYPE_COUNT	//It will be assigned the Number of errors
	};
	//This array should be initialized by calling InitializeStaticData() at the start of the application just once.
	static const TCHAR *ErrorString[ ERROR_TYPE_COUNT ];
	static const TCHAR *ErrorFormatString;
	static void InitializeStaticData();

};
}