#include "NFA.h"
#include "DFA.h"
#include "stdafx.h"
#include "Globals.h"
#include "FAFactory.h"
#include "StreamReader.h"
#include "utf8Codecvt.h"
#include "LexerInputReader.h"
#include "CompressedDFA.h"
#include <algorithm>
#include <locale>
#include <string>
#include <fstream>
#include <cstdio>
using namespace std;
using namespace Automata;

#ifdef _UNICODE
	#define tout	wcout
	#define tin		wcin
	#define terr	wcerr
#else	//ASCII
	#define tout	cout
	#define tin		cin
	#define terr	cerr
#endif

/*****Globals*****/
tstring		inputFileName;
tstring		outputFileName;

InputStream									inputStream;
basic_ofstream<TCHAR, char_traits<TCHAR> >	outputStream;

/*****Functions*****/
//It takes errorsTable and prints the errors.
void PrintStringTable(const vector<tstring> &stringTable);

//It prints the (key, value) pairs of table.
void PrintDictionary(const Dictionary &table);

//It takes an array of strings representing the command line
//arguments, and updates file names and other global variables
//correspondingly. It notifies the user of any errors it faces.
//argv: an array of strings (command line arguments).
//argc: number of entries in the array
//It returns TRUE if the command line arguments are in correct 
//format.
BOOL ParseCommandLineArguments(int argc, _TCHAR *argv[]);

//It tries to open the specified input file. If failed, it returns
//FALSE.
BOOL PrepareInput();

//Fills the option table with the default values for all options.
void InitializeOptionTable(Dictionary &optionTable);
//
void Help();

//Reports information about the DFA
void ReportStatus(const DFA& DFA, const TCHAR *name);

//Reports information about the DFA
void ReportStatus_Old(const DFA& DFA, const TCHAR *name);

double NumberOfKBs(int bytes);

int _tmain(int argc, _TCHAR* argv[])
{
	const int SIZE = 512;
	BOOL		success;
	Dictionary	options;
	DFA			dfa;
	TCHAR		match[SIZE];
	int			strLen;
	char		*fileName;
	vector<tstring> warnings;
	tstring		language;
	tstring		compressionTechnique;
	tstring		temp;
	int			pairsThreshold;
	int			dfaBytes;
	int			cdfaBytes;
	/***************************************************
	Level 1 Errors	: Errors encountered during parsing the file
					and invalid character set specification.
	Level 2 Errors	: Errors due to invalid option values. These
					include:
					1- Language. Valid = {C++, C#}
					2- Compression Technique. Valid = {None, 
					Redundancy,	Pairs, Best}
					3- Compression Threshold: the threshold argument
					specified for the pairs compression. Valid if 
					Compression Technique != Pairs || != Best || > 0.
					4- Output file name. Valid if it is valid file
					name.
					Other errors with options will be reported by
					the compiler of the generated code.
	Warnings		: Not implemented yet.
	****************************************************/
	LexerInputReader::InitializeStaticData();

	//0) Load default options.
	InitializeOptionTable( options );

	//1) Parse command line arguments.
	success = ParseCommandLineArguments( argc, argv );
	if(! success)
		return 1;

	//2) Try openning the input file.
	success = PrepareInput();
	if(! success)
	{
		terr<<_T("Cannot open the specified file: ")<<inputFileName
			<<endl;
		return 1;
	}

	//3) Parse the input file.
	LexerInputReader lexer( inputStream, options, inputFileName );
	lexer.ParseFile();

	//4) Check for level1 errors and terminate if any.
	if(Level1ErrorsTable.size() > 0)
	{
		PrintStringTable( Level1ErrorsTable );
		return 1;
	}
		
	//5) Construct the dfa
	dfa.Reset(& lexer._faFactory, lexer.Nfa, lexer.Code );	
	//Report about the dfa
	ReportStatus( dfa, (inputFileName + _T(" before minimization")).c_str() );
	//6) Minimize the dfa
	dfa.Minimize();
	//Report about the dfa after minimization.
	ReportStatus( dfa, (inputFileName + _T(" after minimization")).c_str() );
	//Save number of bytes
	dfaBytes = dfa.NumberOfBytes();
	//7) Validate and process the options
	outputFileName = options[OPTION_FILE_NAME];
	language = options[OPTION_LANGUAGE];
	compressionTechnique = options[OPTION_COMPRESSION_TECHNIQUE];
	temp = options[OPTION_PAIRS_THRESHOLD];
	//Capitalize language and compressionTechnique
	transform(language.begin(),
			  language.end(),
			  language.begin(),
			  toupper);
	transform(compressionTechnique.begin(),
			  compressionTechnique.end(),
			  compressionTechnique.begin(),
			  toupper);
	pairsThreshold = -10; //Initialize with an invalid value.
	if( compressionTechnique == _T("PAIRS") ||
		compressionTechnique == _T("BEST") )
	{
		pairsThreshold = _ttoi(temp.c_str()); //to integer
		//Check pairs threshold
		if(pairsThreshold < 0)
		{
			TCHAR buffer[1024];

			_stprintf(	buffer,
				_T(" \t%d\tError:%s\t%s\t \t "),
						errCount++,
						_T("Invalid threshold specification for the Pairs compression Technique."),
						_T(" "));

			tstring error( buffer );
			Level2ErrorsTable.push_back(buffer);
		}
	}

	//8) Check language, compression and perform code generation.
	if(language == _T("C++"))
	{
		#ifdef _UNICODE
			//Get ascii names
			strLen = outputFileName.size();
			fileName = new char[ strLen + 1];

			wcstombs( fileName, outputFileName.c_str(),  strLen );
			fileName[strLen] = '\0';

			string asciiFileName = fileName;

			delete[] fileName;

			tofstream tout_cpp((asciiFileName + ".cpp").c_str());
			tofstream tout_h((asciiFileName + ".h").c_str());
		#else
			tofstream tout_cpp((outputFileName + _T(".cpp")).c_str());
			tofstream tout_h((outputFileName + _T(".h")).c_str());
		#endif
		//Check if files were openned successfully.
		if(!tout_h.is_open() || !tout_cpp.is_open())
		{
			TCHAR buffer[1024];

			_stprintf(	buffer,
				_T(" \t%d\tError:%s\t%s\t \t "),
						errCount++,
						_T("Cannot create the specified output files for code generation."),
						_T(" "));

			tstring error( buffer );
			Level2ErrorsTable.push_back(buffer);
		}
		if( compressionTechnique == _T("REDUNDANCY") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA_RedundancyRemoved* cdfa = dfa.CompressToRedundancyRemoved(/*500*/);
			cdfa->GenerateCodeCpp(
					tout_cpp,
					tout_h,
					options,
					lexer.HeaderTopSection,
					lexer.ClassDefinition,
					lexer.CppTopSection,
					lexer.CppBottomSection
					);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;

			delete cdfa;
			tout_cpp.close();
			tout_h.close();
		}
		else if( compressionTechnique == _T("PAIRS") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA_Pairs_Normal* cdfa = dfa.CompressToPairs_Normal(pairsThreshold);
			cdfa->GenerateCodeCpp(
					tout_cpp,
					tout_h,
					options,
					lexer.HeaderTopSection,
					lexer.ClassDefinition,
					lexer.CppTopSection,
					lexer.CppBottomSection
					);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;
			delete cdfa;
			tout_cpp.close();
			tout_h.close();
		}
		else if( compressionTechnique == _T("BEST") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA* cdfa = dfa.MostCompactTechnique(pairsThreshold);
			cdfa->GenerateCodeCpp(
					tout_cpp,
					tout_h,
					options,
					lexer.HeaderTopSection,
					lexer.ClassDefinition,
					lexer.CppTopSection,
					lexer.CppBottomSection
					);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;
			delete cdfa;
			tout_cpp.close();
			tout_h.close();
		}
		else if( compressionTechnique == _T("NONE") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			dfa.GenerateCodeCpp(
				tout_cpp,
				tout_h,
				options,
				lexer.HeaderTopSection,
				lexer.ClassDefinition,
				lexer.CppTopSection,
				lexer.CppBottomSection
				);
			tout_cpp.close();
			tout_h.close();
		}
		else //Unsupported compression
		{
			Level2ErrorsTable.push_back( _T("Invalid compression specification."));
			PrintStringTable( Level2ErrorsTable );
			return 1;
		}
	}
	else if(language == _T("C#"))
	{
		#ifdef _UNICODE
			//Get ascii names
			strLen = outputFileName.size();
			fileName = new char[ strLen + 1];

			wcstombs( fileName, outputFileName.c_str(),  strLen );
			fileName[strLen] = '\0';

			string asciiFileName = fileName;

			delete[] fileName;

			tofstream tout_cs((asciiFileName + ".cs").c_str());
		#else
			//start of code generation
			tofstream tout_cs((outputFileName + _T(".cs")).c_str());
		#endif
		//Check if the file was openned successfully.
		if(!tout_cs.is_open())
		{
			Level2ErrorsTable.push_back(_T("Cannot create the specified output files for code generation."));
		}
		if( compressionTechnique == _T("REDUNDANCY") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA_RedundancyRemoved* cdfa = dfa.CompressToRedundancyRemoved(/*500*/);
			cdfa->GenerateCodeCs(
					tout_cs,
					options,
					lexer.HeaderTopSection,
					lexer.ClassDefinition,
					lexer.CppTopSection,
					lexer.CppBottomSection
					);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;
			delete cdfa;
			tout_cs.close();
		}
		else if( compressionTechnique == _T("PAIRS") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA_Pairs_Normal* cdfa = dfa.CompressToPairs_Normal(pairsThreshold);
			cdfa->GenerateCodeCs(
					tout_cs,
					options,
					lexer.HeaderTopSection,
					lexer.ClassDefinition,
					lexer.CppTopSection,
					lexer.CppBottomSection
					);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;
			delete cdfa;
			tout_cs.close();
		}
		else if( compressionTechnique == _T("BEST") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			CompressedDFA* cdfa = dfa.MostCompactTechnique(pairsThreshold);
			cdfa->GenerateCodeCs(
				tout_cs,
				options,
				lexer.HeaderTopSection,
				lexer.ClassDefinition,
				lexer.CppTopSection,
				lexer.CppBottomSection
				);
			cdfaBytes = cdfa->MemorySize();
			tout<<endl
				<<_T("Number of bytes after compression: ")<<cdfaBytes<<endl
				<<_T("Number of KBs: ")<<NumberOfKBs(cdfaBytes)<<endl
				<<_T("Compression ratio: ")<<100.0 * ((double)cdfaBytes)/dfaBytes<<_T("%")<<endl;
			delete cdfa;
			tout_cs.close();
		}
		else if( compressionTechnique == _T("NONE") )
		{
			if(Level2ErrorsTable.size() > 0)
			{
				PrintStringTable( Level2ErrorsTable );
				return 1;
			}
			dfa.GenerateCodeCs(
				tout_cs,
				options,
				lexer.HeaderTopSection,
				lexer.ClassDefinition,
				lexer.CppTopSection,
				lexer.CppBottomSection
				);
			tout_cs.close();
		}
		else //Unsupported compression
		{
			Level2ErrorsTable.push_back( _T("Invalid compression specification."));
			PrintStringTable( Level2ErrorsTable );
			return 1;
		}
	}
	//else if(language == _T("JAVA"))
	/*{
		#ifdef _UNICODE
		//Get ascii names
		strLen = outputFileName.size();
		fileName = new char[ strLen + 1];

		wcstombs( fileName, outputFileName.c_str(),  strLen );
		fileName[strLen] = '\0';

		string asciiFileName = fileName;

		delete[] fileName;

		tofstream tout_cpp((asciiFileName + ".cpp").c_str());
		tofstream tout_h((asciiFileName + ".h").c_str());
		tofstream tout_cs((asciiFileName + ".cs").c_str());
		tofstream tout_java((asciiFileName + ".java").c_str());

	#else
		//start of code generation
		tofstream tout_cpp((outputFileName + _T(".cpp")).c_str());
		tofstream tout_h((outputFileName + _T(".h")).c_str());
		tofstream tout_cs((outputFileName + _T(".cs")).c_str());
		tofstream tout_java((outputFileName + _T(".java")).c_str());
	#endif
	}*/
	else
	{
		Level2ErrorsTable.push_back(_T("Invalid language specification."));
	}
	
	if(Level2ErrorsTable.size()>0)
	{
		PrintStringTable( Level2ErrorsTable );
		return 1;
	}
	
/*	TCHAR *strrtest = _T("HelloWorld\n");
	do
	{
		tout<<_T("Enter a string to match: ");
//		tin.getline(match, SIZE, _T('\n'));
		if(*match == NULL) break;
		const TCHAR *result = dfa.GetMatchingAction(strrtest);//match);
		tout<<_T("Result: ");
		if(result == NULL)	tout<<_T("No result. It is NULL.");
		else tout<<result;
		tout<<endl;
	}while(*match!=NULL);*/
	return 0;
}

BOOL ParseCommandLineArguments(int argc, _TCHAR *argv[])
{
	if(argc  < 2)
	{
		tout<<_T("The syntax of the command is incorrect.\n");
		tout<<_T("Syntax: ")<<_T("lexer inputFileName [outputFileName].")
			<<endl;
		return FALSE;
	}
	if(argc > 3)
	{
		tout<<_T("The syntax of the command is incorrect.\n");
		tout<<_T("Syntax: ")<<_T("lexer inputFileName [outputFileName].")
			<<endl;
		return FALSE;
	}
	inputFileName = argv[1];
	if( argc == 3 ) outputFileName = argv[2];
	else //inputFileName.
	{
		outputFileName = argv[1];
		int dotPosition = outputFileName.find_last_of( _T('.') );
		outputFileName.erase( dotPosition );
	}
	return TRUE;
}
//////
BOOL PrepareInput()
{
	if(! inputStream.Open(inputFileName.c_str()) ) return FALSE;
	
	//Open now the output file
//	strLen = outputFileName
//	fname = new char [
	return TRUE;
}
//////
void PrintStringTable(const vector<tstring> &stringTable)
{
	vector<tstring>::size_type i = 0, count = stringTable.size();
	//tout<<_T("Number of entries: ")<<count<<endl;
	for(; i<count; i++)
	{
		terr<<stringTable[i]<<endl;
		tout<<stringTable[i]<<endl;
	}
}
//////
void PrintDictionary(const Dictionary &table)
{
	const TCHAR *formatString = _T("%-24s\t%-64s%s");
	Dictionary::const_iterator	iter,end;

	tout<<_T("Number of entries: ")<<table.size()<<endl;
	if(table.size() < 1)
		return;

	_tprintf( formatString, _T("Key"), _T("Value"), ENDLINE);
	iter = table.begin();
	end = table.end();
	for(; iter != end; iter++)
		_tprintf(	formatString,
					iter->first.c_str(),
					iter->second.c_str(),
					ENDLINE);
}
//////
void InitializeOptionTable(Dictionary &optionTable)
{
	optionTable[ OPTION_LANGUAGE ] = _T("C++");
	optionTable[ OPTION_CHARACTER_SET ] = _T("[\\d0-\\d127]");
	optionTable[ OPTION_NAMESPACE ] = _T("Compiler");
	optionTable[ OPTION_FUNCTION_NAME ] = _T("GetNextToken");
	optionTable[ OPTION_RETURN_TYPE ] = _T("int");
	optionTable[ OPTION_CLASS_NAME ] = _T("LexicalAnalyzer");
	optionTable[ OPTION_FILE_NAME ] = _T("lex");
	optionTable[ OPTION_INVALID_TOKEN_ACTION ] = _T("-2");
	optionTable[ OPTION_EOF_ACTION ] = _T("-1");
	optionTable[ OPTION_COMPRESSION_TECHNIQUE ] = _T("Redundancy");
	optionTable[ OPTION_PAIRS_THRESHOLD ] = _T("8");
	optionTable[ OPTION_SUPER_CLASS ] = _T("");
}
//////
void ReportStatus_Old(const DFA& DFA, const TCHAR *name)
{
	tout<<endl<<endl
		<<_T("-------------------------")<<endl
		<<_T("DFA ")<<name<<" status:"<<endl
		<<_T("-------------------------")<<endl
		<<_T("IsRejectionMachine():    ")<<(DFA.IsRejectionMachine()?"TRUE":"FALSE")<<endl
		<<_T("NumberOfStates():        ")<<DFA.NumberOfStates()<<endl
		<<_T("-------------------------")<<endl
		<<endl<<endl;
}
//////
double NumberOfKBs(int bytes)
{
	return ((double)bytes)/1024.0;
}
//////
void ReportStatus(const DFA& DFA, const TCHAR *name)
{
	tout<<endl
		<<_T("-------------------------")<<endl
		<<_T("DFA ")<<name<<" status:"<<endl
		<<_T("-------------------------")<<endl
		<<_T("Number of states:        ")<<DFA.NumberOfStates()<<endl
		<<_T("Number of symbols:       ")<<DFA.NumberOfSymbols()<<endl
		<<_T("Number of bytes:         ")<<DFA.NumberOfBytes()<<endl
		<<_T("Number of KBs:           ")<<NumberOfKBs(DFA.NumberOfBytes())<<endl
		<<_T("-------------------------")//<<endl
		<<endl;
}