#include "stdafx.h"
#include "Globals.h"
#include "FAFactory.h"
#include "NFA.h"
#include "RangedSet.h"
#include "LexerInputReader.h"
#include "RegularExpressionParser.h"
using namespace Automata;
using namespace std;
typedef LexerInputReader Lir;
RegularExpressionTokenizer::RegularExpressionTokenizer(
									LexerInputReader *owner)
									: Lexeme(1024, NULL),
						CharacterClass( owner->_faFactory.NumberOfSymbols() - 1,
												TRUE )

{
	_lexerReader = owner;
	_string = NULL;
	Token = EndOfString;
}
//////
void RegularExpressionTokenizer::Reset( const tstring &regex )
{
	_string = &regex;
	_current = regex.begin();
	Advance();
}
//////
BOOL RegularExpressionTokenizer::Advance()
{
	BOOL			reachedEnd = FALSE;
	BOOL			correctMatch;
	tstrIterator	end = _string->end(), tempIter;
	TCHAR			c;
	Lir::ErrorCode	errCode;
	int				strLen;
	RangedSet		rangedSet( 
						&CharacterClass,
						_lexerReader->_faFactory.startSymbol);
    Lexeme.erase();
	if( _current >= end )  // reached the end.
	{
		Token = EndOfString;
		return FALSE;
	}
	Token = InvalidToken;
	c = * _current;
	switch(c)
	{
	case _T('{'): //{Macro}
		_current++; //Advance
		if( _current >= end )
		{
            _lexerReader->AddError( 
				Lir::INVALID_MACRO_INVOCATION_WITHIN_REGEX);
			Token = InvalidToken;
			return TRUE;
		}
		else if( !_istalpha( *_current ) && *_current != _T('_') )
		{
			_lexerReader->AddError( 
				Lir::INVALID_MACRO_INVOCATION_WITHIN_REGEX);
			Token = InvalidToken;
			return TRUE;
		}
		//So, *_current = [_a-zA-Z]
		tempIter = MatchIdentifier( _current, end );
		//The macroString is [_current, tempIter)
		//But we check that tempIter = }
		if( *tempIter != _T('}') )
		{
			_lexerReader->AddError( 
				Lir::INVALID_MACRO_INVOCATION_WITHIN_REGEX );
			Token = InvalidToken;
			return TRUE;
		}
		//Copy into Lexeme
        Lexeme.assign( _current, tempIter );
		Token = Macro;
		//Update _current to point after {Macro}
		_current = tempIter; //*_current = }
		_current++;
		return TRUE;
	case _T('['): //CCL_BEGIN
		//Check that it is not []
		if( _current+1 >= end ) //[ is the last char
		{
			_lexerReader->AddError( 
				Lir::UNEXPECTED_END_OF_CCL );
			Token = InvalidToken;
			return TRUE;
		}
		if( *( _current + 1 ) == _T(']') ) //[]: Epsilon
		{
			_current += 2;
			Token = Epsilon;
			return TRUE;
		}
		//Nice, big CCL ! [...]
		rangedSet.Clear();
		errCode = Lir::ParseCharacterClass( _current, 
											end, 
											TRUE,
											rangedSet );
		if( errCode == Lir::NO_ERROR )
		{
			Token = CCL;
			return TRUE;
		}
		else
		{
			Token = InvalidToken;
			_lexerReader->AddError( errCode );
			return TRUE;
		}
		break; //useless
	case _T('\\'): //backslash
		if( _current + 1 >= end ) //if ends with backslash
		{
			Token = InvalidToken;
			_lexerReader->AddError(
				Lir::REGEX_TERMINATED_WITH_BACKSLASH );
			return TRUE;
		}
		_current++; //Precond: *(_current - 1 ) = '\\'
		Lexeme.assign(1, ParseEscapeSequence( _current, end ) );
		Token = Symbol;
		return TRUE;
	case _T('\"'): // "Quoted_Text"
		_current++;
		if(_current >= end) //ends with " ?
		{
			Token = InvalidToken;
			_lexerReader->AddError(
				Lir::UNEXPECTED_END_OF_REGEX );
			return TRUE;
		}
		//Advice: In order to understand function calls, see
		//the arguments' names using the text lookup feature.
		correctMatch = MatchQuotedText( _current,
										end,
										TRUE,
										FALSE,
										Lexeme);
		if(! correctMatch ) //Faced errors ?
		{
			Token = InvalidToken;
			_lexerReader->AddError( 
				Lir::UNEXPECTED_END_OF_REGEX );
			return TRUE;
		}
		//Otherwise, we matched "..." successfully.
		//One still last check: Is all the chars are within
		//the define character set.
		strLen = Lexeme.size();
		for(int i=0; i<strLen; i++)
			if(! rangedSet.WithinRange( Lexeme[i] ) )
			{
				Token = InvalidToken;
				_lexerReader->AddError( 
					Lir::CHAR_OUTSIDE_DEFINED_CHARSET );
				return TRUE;
			}
		Token = QuotedText;
		//since _current points to the closing ", increment it
		_current++;
		return TRUE;
	case _T('*'): 
		Token = Star;
		_current++;
		return TRUE;
	case _T('|'):
		Token = OR;
		_current++;
		return TRUE;
	case _T('+'):
		Token = Plus;
		_current++;
		return TRUE;
	case _T('?'):
		Token = QuestionMark;
		_current++;
		return TRUE;
	case _T('.'):
		Token = Dot;
		_current++;
		return TRUE;
	case _T('('):
		Token = LeftBracket;
		_current++;
		return TRUE;
	case _T(')'):
		Token = RightBracket;
		_current++;
		return TRUE;
	default: //Any other char ?
		//Check wether within the defined character set or not
		if(! rangedSet.WithinRange( c ) )
		{
			Token = InvalidToken;
			_lexerReader->AddError( 
					Lir::CHAR_OUTSIDE_DEFINED_CHARSET );
			return TRUE;
		}
		Lexeme = c;
		Token = Symbol;
		_current++;
		return TRUE;
	}
	// IMPOSSIBLE
}
//////
BOOL RegularExpressionTokenizer::Match( 
								RegularExpressionToken token )
{
	if( token != Token )
		return FALSE;
	Advance();
	return TRUE;
}
//////
RegularExpressionParser::RegularExpressionParser(
							LexerInputReader *owner) :
										_lexerReader(owner),
										_scanner( owner )
{
	//Save the address of the memory manager once and for all.
	_memManager = & (_lexerReader->_mem);	
}
//////
RegularExpressionParser::~RegularExpressionParser(void)
{
}
//////
BOOL RegularExpressionParser::Parse(const tstring &regex,
									NFA &nfa)
{
	_scanner.Reset( regex );
	//_memManager = nfa.memory; Already Saved
	return Regex( nfa );
}

//Regex ---> ConcateRegex { OR ConcateRegex }
BOOL RegularExpressionParser::Regex(NFA &nfa)
{
	BOOL	retval;
	NFA		tnfa( _memManager ); //temp nfa
	retval = ConcateRegex( nfa );
	if( !retval ) return retval; //FALSE
	while( _scanner.Match( OR ) ) //{ OR CRg}
	{
		retval = ConcateRegex( tnfa ); //CRg
		if( ! retval ) return retval; //FALSE
		nfa.Or( tnfa );
	}
	return retval; //TRUE
}

//ConcateRegex ---> Term { Term }
//First(Term) = { QText, Symbol, CCL, Epsilon, Macro, (, Dot }
BOOL RegularExpressionParser::ConcateRegex(NFA &nfa)
{
	BOOL	retval;
	NFA		tnfa( _memManager ); //temp nfa

	retval = Term( nfa );		//Term
	if( ! retval ) return retval; //FALSE
	
	while(true)
	{
		switch( _scanner.Token ) //token belongs to First(Term) ?
		{
		case QuotedText:
		case Symbol:
		case CCL:
		case Epsilon:
		case Macro:
		case LeftBracket:
		case Dot:
			retval = Term( tnfa ); //The other term
			if(! retval) return retval; //failed ?
			nfa.Concatenate( tnfa ); //Concatenate
			break;
		default:
			return retval; //TRUE
		}
	}
}

//Term ---> Atom { Plus | Star | QuestionMark }
BOOL RegularExpressionParser::Term(NFA &nfa)
{
	if( ! Atom(nfa) ) 
		return FALSE;
	while(true)
	{
		switch( _scanner.Token )
		{
		case Star:
			_scanner.Advance();
			nfa.Closure();
			break;
		case Plus:
			_scanner.Advance();
			nfa.PlusClosure();
			break;
		case QuestionMark:
			_scanner.Advance();
			nfa.OptionalClosure();
			break;
		default:
			return TRUE;
		}
	}
}

//Atom ---> QuotedText | Symbol | . | Epsilon | CCL | Macro
//				| ( Regex )
BOOL RegularExpressionParser::Atom(NFA &nfa)
{
	NFA		tnfa( _memManager );
	switch( _scanner.Token ) //which case is that
	{
	case QuotedText: //"...."
		{
			int strLen = _scanner.Lexeme.size();
			if(strLen >= 1)
				nfa.ReallocateAndSet( _scanner.Lexeme[0] );
			for(int i=1; i<strLen; i++) //make nfas and concatenate
										//them
			{
				tnfa.ReallocateAndSet( _scanner.Lexeme[i] );
				nfa.Concatenate( tnfa );
			}
			_scanner.Advance();
			return TRUE;
		}
	case Symbol:
		nfa.ReallocateAndSet( _scanner.Lexeme[0] );
		_scanner.Advance();
		return TRUE;
	case Dot: //Every char except endline
		{
			const FAFactory *charSet = _memManager->GetFAFactory();
			BitSet *allExceptEndline = new BitSet
				( charSet->NumberOfSymbols() - 1,
					FALSE );
			//Check if charSet includes '\n' or not
			int cindex = charSet->IndexOf( _T('\n') );
			if(cindex != -1)	//If it exists, remove it. Else, already
								//removed
				allExceptEndline->Remove( cindex );
			nfa.SetCharacterClass( allExceptEndline );
			_scanner.Advance();
			return TRUE;
		}
	case Epsilon:
		nfa.SetAcceptEmptyWord();
		_scanner.Advance();
		return TRUE;
	case CCL:
		nfa.SetCharacterClass(new BitSet( _scanner.CharacterClass ) );
		_scanner.Advance();
		return TRUE;
	case Macro: //Get the macro value and parse it.
		{
			Dictionary::iterator iter = 
				_lexerReader->_macroTable.find( _scanner.Lexeme );
			_scanner.Advance();
			if( iter != _lexerReader->_macroTable.end() ) //found ?
			{
				//Construct a new parser
				RegularExpressionParser macroParse( _lexerReader );
				//Parse the macro value
				return macroParse.Parse( iter->second, nfa );
			}
			//else, undefined macro
			_lexerReader->AddError( Lir::INVOKED_MACRO_IS_UNDEFINED );
			return FALSE;
		}
	case LeftBracket: //( Regex )
		_scanner.Advance();
		return Regex( nfa ) && _scanner.Match( RightBracket );
	default: //What ?
		return FALSE;
	}
}