#pragma once
#include "stdafx.h"
#include "LexerInputReader.h"
#include "FAFactory.h"
#include "NFA.h"

namespace Automata
{
	enum RegularExpressionToken
	{
		InvalidToken,
		OR,	//Vertical Bar: |
		Macro,	//{Identifier}
		CCL,	//[...]
		QuotedText, //"..."
		Star,	//*
		Plus,	//+
		QuestionMark, //?
		Dot,	//.
		Symbol, //any thing else
		Epsilon, //Empty word
		LeftBracket, // (
		RightBracket,// )
		EndOfString //EOS
	};
	//The lexer reader object must have FAFactory properly 
	//initialized.
	class RegularExpressionTokenizer
	{
	private:
        LexerInputReader	*_lexerReader; //Owner reader object
		const tstring		*_string; //Regular expression string
		tstrIterator		_current; //current location in _string
	public:
		tstring					Lexeme; //The current lexeme
		RegularExpressionToken	Token; //The current token
		BitSet					CharacterClass;

		RegularExpressionTokenizer(LexerInputReader *owner);

		//Resets the tokenizer for a new regex. The function
		//operates on the supplied string without changin it.
		//So, it must remains in memory during the lexical analysis.
		void Reset( const tstring &regex );

		//Advances the reading position and matches a new token.
		//It returns FALSE if we reached the end of the string.
		BOOL Advance();

		//Sees if the current token is the same as the parameter
		//token. It returns TRUE if they are the same and 
		//advances the reading position by reading a new token.
		BOOL Match(RegularExpressionToken token);
	};
	/*
	It parses regular expression and performs Thompson
	construction on the run and produces an equivalent NFA.
	*/
	class RegularExpressionParser
	{
	private:
		//Owner reader object
		LexerInputReader	*_lexerReader;

		//Scanner
		RegularExpressionTokenizer _scanner;

		//This is put here for reducing the stack memory
		//needed for holding it in the recursive functions.
		NFA::MemoryManager	*_memManager;
	public:
		RegularExpressionParser(LexerInputReader *owner);

		//It tries to parse the supplied regular expression regex.
		//If the parse is successful, it constructs the 
		//corresponding NFA starting at nfa, and the function returns 
		//TRUE. If the parse fails, it returns FALSE and adds the
		//relevant errors to the ErrorTable in the owner
		//LexerInputReader object.
		BOOL	Parse(	const tstring &regex,
						NFA &nfa);
		~RegularExpressionParser(void);
	private:
		/***Recursive Descent Parsing Functions***/
		
		//It parses the regular expression and constructs the 
		//corresponding NFA into nfa.
		//Regex ---> ConcateRegex { OR ConcateRegex }
		BOOL	Regex(NFA &nfa);
		
		//ConcateRegex ---> Term { Term }
		BOOL	ConcateRegex(NFA &nfa);

		//Term ---> Atom { Plus | Star | QuestionMark }
		BOOL	Term(NFA &nfa);

		//Atom ---> QuotedText | Symbol | . | Epsilon | CCL | Macro
		//				| ( Regex )
		BOOL	Atom(NFA &nfa);
	};
}